from __future__ import annotations

import collections
import datetime as dt
import functools
import inspect
import numbers
import os
import re
import traceback
import warnings
from collections import OrderedDict, abc, defaultdict
from contextlib import contextmanager
from numbers import Real
from textwrap import dedent
from threading import get_ident
from typing import TYPE_CHECKING, Callable, TypeVar

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec

    P = ParamSpec("P")
    R = TypeVar("R")
    CallableT = TypeVar("CallableT", bound=Callable)

DEFAULT_SIGNATURE = inspect.Signature([
    inspect.Parameter('self', inspect.Parameter.POSITIONAL_OR_KEYWORD),
    inspect.Parameter('params', inspect.Parameter.VAR_KEYWORD),
])

MUTABLE_TYPES = (abc.MutableSequence, abc.MutableSet, abc.MutableMapping)

class ParamWarning(Warning):
    """Base Param Warning."""

class ParamPendingDeprecationWarning(ParamWarning, PendingDeprecationWarning):
    """
    Param PendingDeprecationWarning.

    This warning type is useful when the warning is not meant to be displayed
    to REPL/notebooks users, as DeprecationWarning are displayed when triggered
    by code in __main__ (__name__ == '__main__' in a REPL).
    """


class ParamDeprecationWarning(ParamWarning, DeprecationWarning):
    """
    Param DeprecationWarning.

    Ignored by default except when triggered by code in __main__
    """


class ParamFutureWarning(ParamWarning, FutureWarning):
    """
    Param FutureWarning.

    Always displayed.
    """


class Skip(Exception):
    """Exception that allows skipping an update when resolving a reference.

    References
    ----------
    https://param.holoviz.org/user_guide/References.html#skipping-reference-updates

    Examples
    --------
    >>> import param
    >>> class W(param.Parameterized):
    ...    a = param.Number()
    ...    b = param.Number(allow_refs=True)
    ...    run = param.Event()
    >>> w = W(a=0, b=2)

    Let's define a function:

    >>> def add(a, b, run):
    ...    if not run:
    ...        raise param.Skip
    ...    return a + b

    Let's use the function as a reference:

    >>> v = W(b=param.bind(add, w.param.a, w.param.b, w.param.run))

    We can see that b has not yet been resolved:

    >>> v.b
    0.0

    Let's trigger `w.param.run` and check `v.b` has been resolved:

    >>> w.param.trigger('run')
    >>> v.b
    2
    """


def _deprecated(extra_msg="", warning_cat=ParamDeprecationWarning):
    def decorator(func):
        """Mark a function or method as deprecated.

        This internal decorator issues a warning when the decorated function
        or method is called, indicating that it has been deprecated and will
        be removed in a future version.

        Parameters
        ----------
        func: FunctionType | MethodType
            The function or method to mark as deprecated.

        Returns
        -------
        callable:
            The wrapped function that issues a deprecation warning.
        """
        @functools.wraps(func)
        def inner(*args, **kwargs):
            msg = f"{func.__name__!r} has been deprecated and will be removed in a future version."
            if extra_msg:
                em = dedent(extra_msg)
                em = em.strip().replace('\n', ' ')
                msg = msg + ' ' + em
            warnings.warn(msg, category=warning_cat, stacklevel=_find_stack_level())
            return func(*args, **kwargs)
        return inner
    return decorator


# Copy of Python 3.2 reprlib's recursive_repr but allowing extra arguments
def _recursive_repr(fillvalue='...'):
    """Decorate a repr function to return a fill value for recursive calls."""

    def decorating_function(user_function):
        repr_running = set()

        def wrapper(self, *args, **kwargs):
            key = id(self), get_ident()
            if key in repr_running:
                return fillvalue
            repr_running.add(key)
            try:
                result = user_function(self, *args, **kwargs)
            finally:
                repr_running.discard(key)
            return result
        return wrapper

    return decorating_function


def _is_auto_name(class_name, instance_name):
    return re.match('^'+class_name+'[0-9]{5}$', instance_name)


def _find_pname(pclass):
    """
    Go up the stack and attempt to find a Parameter declaration of the form
    `pname = param.Parameter(` or `pname = pm.Parameter(`.
    """
    stack = traceback.extract_stack()
    for frame in stack:
        match = re.match(r"^(\S+)\s*=\s*(param|pm)\." + pclass + r"\(", frame.line)
        if match:
            return match.group(1)


def _validate_error_prefix(parameter, attribute=None):
    """
    Generate an error prefix suitable for Parameters when they raise a validation
    error.

    - unbound and name can't be found: "Number parameter"
    - unbound and name can be found: "Number parameter 'x'"
    - bound parameter: "Number parameter 'P.x'"
    """
    from param.parameterized import ParameterizedMetaclass

    pclass = type(parameter).__name__
    if parameter.owner is not None:
        if issubclass(type(parameter.owner), ParameterizedMetaclass):
            powner = parameter.owner.__name__
        else:
            powner = type(parameter.owner).__name__
    else:
        powner = None
    pname = parameter.name
    out = []
    if attribute:
        out.append(f'Attribute {attribute!r} of')
    out.append(f'{pclass} parameter')
    if pname:
        if powner:
            desc = f'{powner}.{pname}'
        else:
            desc = pname
        out.append(f'{desc!r}')
    else:
        try:
            pname = _find_pname(pclass)
            if pname:
                out.append(f'{pname!r}')
        except Exception:
            pass
    return ' '.join(out)

def _is_mutable_container(value):
    """Determine if the value is a mutable container.

    Mutable containers typically require special handling when being copied.
    """
    return isinstance(value, MUTABLE_TYPES)

def full_groupby(l, key=lambda x: x):
    """Groupby implementation which does not require a prior sort."""
    d = defaultdict(list)
    for item in l:
        d[key(item)].append(item)
    return d.items()

def iscoroutinefunction(function):
    """Whether the function is an asynchronous generator or a coroutine."""
    # Partial unwrapping not required starting from Python 3.11.0
    # See https://github.com/holoviz/param/pull/894#issuecomment-1867084447
    while isinstance(function, functools.partial):
        function = function.func
    return (
        inspect.isasyncgenfunction(function) or
        inspect.iscoroutinefunction(function)
    )

async def _to_async_gen(sync_gen):
    import asyncio

    done = object()

    def safe_next():
        # Converts StopIteration to a sentinel value to avoid:
        # TypeError: StopIteration interacts badly with generators and cannot be raised into a Future
        try:
            return next(sync_gen)
        except StopIteration:
            return done

    while True:
        value = await asyncio.to_thread(safe_next)
        if value is done:
            break
        yield value

def flatten(line):
    """
    Flatten an arbitrarily nested sequence.

    Inspired by: pd.core.common.flatten

    Parameters
    ----------
    line : sequence
        The sequence to flatten

    Notes
    -----
    This only flattens list, tuple, and dict sequences.

    Returns
    -------
    flattened : generator

    """
    for element in line:
        if any(isinstance(element, tp) for tp in (list, tuple, dict)):
            yield from flatten(element)
        else:
            yield element


def accept_arguments(
    f: Callable[Concatenate[CallableT, P], R]
) -> Callable[P, Callable[[CallableT], R]]:
    """Decorate a decorator to accept arguments."""
    @functools.wraps(f)
    def _f(*args: P.args, **kwargs: P.kwargs) -> Callable[[CallableT], R]:
        return lambda actual_f: f(actual_f, *args, **kwargs)
    return _f


def _produce_value(value_obj):
    """Produce an actual value from a stored object.

    If the object is callable, call it; otherwise, return the object.
    """
    if callable(value_obj):
        return value_obj()
    else:
        return value_obj


def _hashable(x):
    """
    Return a hashable version of the given object x, with lists and
    dictionaries converted to tuples.  Allows mutable objects to be
    used as a lookup key in cases where the object has not actually
    been mutated. Lookup will fail (appropriately) in cases where some
    part of the object has changed.  Does not (currently) recursively
    replace mutable subobjects.
    """
    if isinstance(x, collections.abc.MutableSequence):
        return tuple(x)
    elif isinstance(x, collections.abc.MutableMapping):
        return tuple([(k,v) for k,v in x.items()])
    else:
        return x


def _named_objs(objlist, namesdict=None):
    """
    Given a list of objects, returns a dictionary mapping from
    string name for the object to the object itself. Accepts
    an optional name,obj dictionary, which will override any other
    name if that item is present in the dictionary.
    """
    objs = OrderedDict()

    objtoname = {}
    unhashables = []
    if namesdict is not None:
        for k, v in namesdict.items():
            try:
                objtoname[_hashable(v)] = k
            except TypeError:
                unhashables.append((k, v))

    for obj in objlist:
        if objtoname and _hashable(obj) in objtoname:
            k = objtoname[_hashable(obj)]
        elif any(obj is v for (_, v) in unhashables):
            k = [k for (k, v) in unhashables if v is obj][0]
        elif hasattr(obj, "name"):
            k = obj.name
        elif hasattr(obj, '__name__'):
            k = obj.__name__
        else:
            k = str(obj)
        objs[k] = obj
    return objs


def _get_min_max_value(min, max, value=None, step=None):
    """Return min, max, value given input values with possible None."""
    # Either min and max need to be given, or value needs to be given
    if value is None:
        if min is None or max is None:
            raise ValueError(
                f'unable to infer range, value from: ({min}, {max}, {value})'
            )
        diff = max - min
        value = min + (diff / 2)
        # Ensure that value has the same type as diff
        if not isinstance(value, type(diff)):
            value = min + (diff // 2)
    else:  # value is not None
        if not isinstance(value, Real):
            raise TypeError('expected a real number, got: %r' % value)
        # Infer min/max from value
        if value == 0:
            # This gives (0, 1) of the correct type
            vrange = (value, value + 1)
        elif value > 0:
            vrange = (-value, 3*value)
        else:
            vrange = (3*value, -value)
        if min is None:
            min = vrange[0]
        if max is None:
            max = vrange[1]
    if step is not None:
        # ensure value is on a step
        tick = int((value - min) / step)
        value = min + tick * step
    if not min <= value <= max:
        raise ValueError(f'value must be between min and max (min={min}, value={value}, max={max})')
    return min, max, value


def _deserialize_from_path(ext_to_routine, path, type_name):
    """
    Call deserialization routine with path according to extension.
    ext_to_routine should be a dictionary mapping each supported
    file extension to a corresponding loading function.
    """
    if not os.path.isfile(path):
        raise FileNotFoundError(
            "Could not parse file '{}' as {}: does not exist or is not a file"
            "".format(path, type_name))
    root, ext = os.path.splitext(path)
    if ext in {'.gz', '.bz2', '.xz', '.zip'}:
        # A compressed type. We'll assume the routines can handle such extensions
        # transparently (if not, we'll fail later)
        ext = os.path.splitext(root)[1]
    # FIXME(sdrobert): try...except block below with "raise from" might be a good idea
    # once py2.7 support is removed. Provides error + fact that failure occurred in
    # deserialization
    if ext in ext_to_routine:
        return ext_to_routine[ext](path)
    raise ValueError(
        "Could not parse file '{}' as {}: no deserialization method for files with "
        "'{}' extension. Supported extensions: {}"
        "".format(path, type_name, ext, ', '.join(sorted(ext_to_routine))))


def _is_number(obj):
    if isinstance(obj, numbers.Number): return True
    # The extra check is for classes that behave like numbers, such as those
    # found in numpy, gmpy2, etc.
    elif (hasattr(obj, '__int__') and hasattr(obj, '__add__')): return True
    # This is for older versions of gmpy
    elif hasattr(obj, 'qdiv'): return True
    else: return False


def _is_abstract(class_: type) -> bool:
    if inspect.isabstract(class_):
        return True
    try:
        return class_.abstract
    except AttributeError:
        return False


def descendents(class_: type, concrete: bool = False) -> list[type]:
    """
    Return a list of all descendent classes of a given class.

    This function performs a breadth-first traversal of the class hierarchy,
    collecting all subclasses of ``class_``. The result includes ``class_`` itself
    and all of its subclasses. If ``concrete=True``, abstract base classes
    are excluded from the result, including :class:`Parameterized` abstract
    classes declared with ``__abstract = True``.

    Parameters
    ----------
    class_ : type
        The base class whose descendants should be found.
    concrete : bool, optional
        If ``True``, exclude abstract classes from the result. Default is ``False``.

        .. versionadded:: 2.3.0

        Added to encourage users to use :func:`descendents` in favor of
        :func:`concrete_descendents` that clobbers classes sharing the same name.

    Returns
    -------
    list[type]
        A list of descendent classes, ordered from the most base to the most derived.

    Examples
    --------
    >>> class A: pass
    >>> class B(A): pass
    >>> class C(A): pass
    >>> class D(B): pass
    >>> descendents(A)
    [A, B, C, D]
    """
    if not isinstance(class_, type):
        raise TypeError(f"descendents expected a class object, not {type(class_).__name__}")
    q = [class_]
    out = []
    while len(q):
        x = q.pop(0)
        out.insert(0, x)
        try:
            subclasses = x.__subclasses__()
        except TypeError:
            # TypeError raised when __subclasses__ is called on unbound methods,
            # on `type` for example.
            continue
        for b in subclasses:
            if b not in q and b not in out:
                q.append(b)
    return [kls for kls in out if not (concrete and _is_abstract(kls))][::-1]


# Could be a method of ClassSelector.
def concrete_descendents(parentclass: type) -> dict[str, type]:
    """
    Return a dictionary containing all subclasses of the specified
    parentclass, including the parentclass (prefer :func:`descendents`).

    Only classes that are defined in scripts that have been run or modules
    that have been imported are included, so the caller will usually first
    do ``from package import *``.

    Only non-abstract classes will be included.

    .. warning::
       ``concrete_descendents`` overrides descendents that share the same
       class name. To avoid this, use :func:`descendents` with ``concrete=True``.
    """
    # Warns
    # -----
    # ParamWarning
    #     ``concrete_descendents`` overrides descendents that share the same
    #     class name. To avoid this, use :func:`descendents` with ``concrete=True``.

    desc = descendents(parentclass, concrete=True)
    concrete_desc = {c.__name__: c for c in desc}
    # Descendents with the same name are clobbered.
    # if len(desc) != len(concrete_desc):
    #     class_count = Counter([kls.__name__ for kls in desc])
    #     clobbered = [kls for kls, count in class_count.items() if count > 1]
    #     warnings.warn(
    #         '`concrete_descendents` overrides descendents that share the same '
    #         'class name. Other descendents with the same name as the following '
    #         f'classes exist but were not returned: {clobbered!r}\n'
    #         'Use `descendents(parentclass, concrete=True)` instead.',
    #         ParamWarning,
    #     )
    return concrete_desc


def _abbreviate_paths(pathspec,named_paths):
    """
    Given a dict of (pathname,path) pairs, removes any prefix shared by all pathnames.
    Helps keep menu items short yet unambiguous.
    """
    from os.path import commonprefix, dirname, sep

    prefix = commonprefix([dirname(name)+sep for name in named_paths.keys()]+[pathspec])
    return OrderedDict([(name[len(prefix):],path) for name,path in named_paths.items()])


def _to_datetime(x):
    """Convert a date object to a datetime object for comparison.

    This internal function ensures that date and datetime objects can be
    compared without errors by converting date objects to datetime objects.

    Parameters
    ----------
    x: Any
        The object to convert, which may be a `date` or `datetime` object.

    Returns
    -------
    datetime.datetime:
        A datetime object if the input was a date object;
        otherwise, the input is returned unchanged.
    """
    if isinstance(x, dt.date) and not isinstance(x, dt.datetime):
        return dt.datetime(*x.timetuple()[:6])
    return x


@contextmanager
def exceptions_summarized():
    """Context manager to display exceptions concisely without tracebacks.

    This utility is useful for writing documentation or examples where
    only the exception type and message are needed, without the full
    traceback.

    Yields
    ------
    None:
        Allows the code inside the context to execute.

    Prints
    ------
    A concise summary of any exception raised, including the exception
    type and message, to `sys.stderr`.
    """
    try:
        yield
    except Exception:
        import sys
        etype, value, tb = sys.exc_info()
        print(f"{etype.__name__}: {value}", file=sys.stderr)


def _in_ipython():
    try:
        get_ipython
        return True
    except NameError:
        return False

_running_tasks = set()

def async_executor(func):
    import asyncio
    try:
        event_loop = asyncio.get_running_loop()
    except RuntimeError:
        event_loop = asyncio.new_event_loop()
    if event_loop.is_running():
        task = asyncio.ensure_future(func())
        _running_tasks.add(task)
        task.add_done_callback(_running_tasks.discard)
    else:
        event_loop.run_until_complete(func())

class _GeneratorIsMeta(type):
    def __instancecheck__(cls, inst):
        return isinstance(inst, tuple(cls.types()))

    def __subclasscheck__(cls, sub):
        return issubclass(sub, tuple(cls.types()))

    def __iter__(cls):
        yield from cls.types()

class _GeneratorIs(metaclass=_GeneratorIsMeta):
    @classmethod
    def __iter__(cls):
        yield from cls.types()

def gen_types(gen_func):
    """Decorate a generator function to support type checking.

    This decorator modifies a generator function that yields different types
    so that it can be used with `isinstance` and `issubclass`.

    Parameters
    ----------
    gen_func : GeneratorType
        The generator function to decorate.

    Returns
    -------
    type:
        A new type that supports `isinstance` and `issubclass` checks
        based on the generator function's yielded types.

    Raises
    ------
    TypeError: If the provided function is not a generator function.
    """
    if not inspect.isgeneratorfunction(gen_func):
        msg = "gen_types decorator can only be applied to generator"
        raise TypeError(msg)
    return type(gen_func.__name__, (_GeneratorIs,), {"types": staticmethod(gen_func)})


def _find_stack_level() -> int:
    """
    Find the first place in the stack that is not inside numbergen and param.

    Inspired by: pandas.util._exceptions.find_stack_level.
    """
    import numbergen
    import param

    ng_dir = os.path.dirname(numbergen.__file__)
    param_dir = os.path.dirname(param.__file__)

    # https://stackoverflow.com/questions/17407119/python-inspect-stack-is-slow
    frame = inspect.currentframe()
    try:
        n = 0
        while frame:
            filename = inspect.getfile(frame)
            if filename.startswith((ng_dir, param_dir)):
                frame = frame.f_back
                n += 1
            else:
                break
    finally:
        # See note in
        # https://docs.python.org/3/library/inspect.html#inspect.Traceback
        del frame
    return n



"""Callable objects that generate numbers according to different distributions."""

import random
import operator
import hashlib
import struct
import fractions

from ctypes import c_size_t
from math import e,pi

import param


from param import __version__

class TimeAware(param.Parameterized):
    """
    Class of objects that have access to a global time function
    and have the option of using it to generate time-dependent values
    as necessary.

    In the simplest case, an object could act as a strict function of
    time, returning the current time transformed according to a fixed
    equation.  Other objects may support locking their results to a
    timebase, but also work without time.  For instance, objects with
    random state could return a new random value for every call, with
    no notion of time, or could always return the same value until the
    global time changes.  Subclasses should thus provide an ability to
    return a time-dependent value, but may not always do so.
    """

    time_dependent = param.Boolean(default=False,  doc="""
       Whether the given time_fn should be used to constrain the
       results generated.""")

    time_fn = param.Callable(default=param.Dynamic.time_fn, doc="""
        Callable used to specify the time that determines the state
        and return value of the object, if time_dependent=True.""")


    def __init__(self, **params):
        super().__init__(**params)
        self._check_time_fn()


    def _check_time_fn(self, time_instance=False):
        """
        If time_fn is the global time function supplied by
        param.Dynamic.time_fn, make sure Dynamic parameters are using
        this time function to control their behaviour.

        If time_instance is True, time_fn must be a param.Time instance.
        """
        if time_instance and not isinstance(self.time_fn, param.Time):
            raise AssertionError("%s requires a Time object"
                                 % self.__class__.__name__)

        if self.time_dependent:
            global_timefn = self.time_fn is param.Dynamic.time_fn
            if global_timefn and not param.Dynamic.time_dependent:
                raise AssertionError("Cannot use Dynamic.time_fn as"
                                     " parameters are ignoring time.")


class TimeDependent(TimeAware):
    """
    Objects that have access to a time function that determines the
    output value. As a function of time, this type of object should
    allow time values to be randomly jumped forwards or backwards,
    but for a given time point, the results should remain constant.

    The time_fn must be an instance of param.Time, to ensure all the
    facilities necessary for safely navigating the timeline are
    available.
    """

    time_dependent = param.Boolean(default=True, readonly=True, doc="""
       Read-only parameter that is always True.""")

    def _check_time_fn(self):
        super()._check_time_fn(time_instance=True)



class NumberGenerator(param.Parameterized):
    """
    Abstract base class for any object that when called produces a number.

    Primarily provides support for using NumberGenerators in simple
    arithmetic expressions, such as abs((x+y)/z), where x,y,z are
    NumberGenerators or numbers.
    """

    def __call__(self):
        raise NotImplementedError

    # Could define any of Python's operators here, esp. if they have operator or ufunc equivalents
    def __add__      (self,operand): return BinaryOperator(self,operand,operator.add)
    def __sub__      (self,operand): return BinaryOperator(self,operand,operator.sub)
    def __mul__      (self,operand): return BinaryOperator(self,operand,operator.mul)
    def __mod__      (self,operand): return BinaryOperator(self,operand,operator.mod)
    def __pow__      (self,operand): return BinaryOperator(self,operand,operator.pow)
    def __div__      (self,operand): return BinaryOperator(self,operand,operator.div)
    def __truediv__  (self,operand): return BinaryOperator(self,operand,operator.truediv)
    def __floordiv__ (self,operand): return BinaryOperator(self,operand,operator.floordiv)

    def __radd__     (self,operand): return BinaryOperator(self,operand,operator.add,True)
    def __rsub__     (self,operand): return BinaryOperator(self,operand,operator.sub,True)
    def __rmul__     (self,operand): return BinaryOperator(self,operand,operator.mul,True)
    def __rmod__     (self,operand): return BinaryOperator(self,operand,operator.mod,True)
    def __rpow__     (self,operand): return BinaryOperator(self,operand,operator.pow,True)
    def __rdiv__     (self,operand): return BinaryOperator(self,operand,operator.div,True)
    def __rtruediv__ (self,operand): return BinaryOperator(self,operand,operator.truediv,True)
    def __rfloordiv__(self,operand): return BinaryOperator(self,operand,operator.floordiv,True)

    def __neg__ (self): return UnaryOperator(self,operator.neg)
    def __pos__ (self): return UnaryOperator(self,operator.pos)
    def __abs__ (self): return UnaryOperator(self,operator.abs)


operator_symbols = {
    operator.add:'+',
    operator.sub:'-',
    operator.mul:'*',
    operator.mod:'%',
    operator.pow:'**',
    operator.truediv:'/',
    operator.floordiv:'//',
    operator.neg:'-',
    operator.pos:'+',
    operator.abs:'abs',
}

def pprint(x, *args, **kwargs):
    """Pretty-print the provided item, translating operators to their symbols."""
    return x.pprint(*args, **kwargs) if hasattr(x,'pprint') else operator_symbols.get(x, repr(x))


class BinaryOperator(NumberGenerator):
    """
    Applies any binary operator to NumberGenerators or numbers to yield a NumberGenerator.

    Parameters
    ----------
    lhs: NumberGenerator or Number
        The left-hand side operand, which can be a NumberGenerator or a number.
    rhs: NumberGenerator or Number
        The right-hand side operand, which can be a NumberGenerator or a number.
    operator :  callable
        The binary operator to apply to the operands.
    reverse : bool, optional
        If `True`, swaps the left and right operands. Defaults to `False`.
    **args:
        Optional keyword arguments to pass to the operator when it is called.

    Notes
    -----
    It is currently not possible to set parameters in the superclass during
    initialization because `**args` is used by this class itself.
    """

    def __init__(self,lhs, rhs, operator, reverse=False, **args):
        super().__init__()
        if reverse:
            self.lhs=rhs
            self.rhs=lhs
        else:
            self.lhs=lhs
            self.rhs=rhs
        self.operator=operator
        self.args=args

    def __call__(self):
        return self.operator(self.lhs() if callable(self.lhs) else self.lhs,
                             self.rhs() if callable(self.rhs) else self.rhs, **self.args)

    def pprint(self, *args, **kwargs):
        return (pprint(self.lhs,      *args, **kwargs) +
                pprint(self.operator, *args, **kwargs) +
                pprint(self.rhs,      *args, **kwargs))


class UnaryOperator(NumberGenerator):
    """
    Applies any unary operator to a NumberGenerator to yield another NumberGenerator.

    Parameters
    ----------
    operand : NumberGenerator
        The NumberGenerator to which the operator is applied.
    operator : callable
        The unary operator to apply to the operand.
    **args:
        Optional keyword arguments to pass to the operator when it is called.

    Notes
    -----
    It is currently not possible to set parameters in the superclass during
    initialization because `**args` is used by this class itself.
    """

    def __init__(self, operand, operator, **args):
        super().__init__()

        self.operand=operand
        self.operator=operator
        self.args=args

    def __call__(self):
        return self.operator(self.operand(),**self.args)

    def pprint(self, *args, **kwargs):
        return (pprint(self.operator, *args, **kwargs) + '(' +
                pprint(self.operand,  *args, **kwargs) + ')')


class Hash:
    """
    A platform- and architecture-independent hash function (unlike
    Python's inbuilt hash function) for use with an ordered collection
    of rationals or integers.

    The supplied name sets the initial hash state.  The output from
    each call is a 32-bit integer to ensure the value is a regular
    Python integer (and not a Python long) on both 32-bit and 64-bit
    platforms. This can be important to seed Numpy's random number
    generator safely (a bad Numpy bug!).

    The number of inputs (integer or rational numbers) to be supplied
    for __call__ must be specified in the constructor and must stay
    constant across calls.
    """

    def __init__(self, name, input_count):
        self.name = name
        self.input_count = input_count
        self._digest = hashlib.md5()
        self._digest.update(name.encode())
        self._hash_struct = struct.Struct( "!" +" ".join(["I"] * (input_count * 2)))


    def _rational(self, val):
        """Convert the given value to a rational, if necessary."""
        I32 = 4294967296 # Maximum 32 bit unsigned int (i.e. 'I') value
        if isinstance(val, int):
            numer, denom = val, 1
        elif isinstance(val, fractions.Fraction):
            numer, denom = val.numerator, val.denominator
        elif hasattr(val, 'numerator') and hasattr(val, 'denominator'):
            # gmpy2 mpq objects have these attributes
            numer, denom = val.numerator, val.denominator
        elif hasattr(val, 'numer'):
            # I think this branch supports gmpy (i.e. not gmpy2)
            (numer, denom) = (int(val.numer()), int(val.denom()))
        else:
            param.main.param.log(param.WARNING, "Casting type '%s' to Fraction.fraction"
                               % type(val).__name__)
            frac = fractions.Fraction(str(val))
            numer, denom = frac.numerator, frac.denominator
        return numer % I32, denom % I32

    def __getstate__(self):
        """Avoid Hashlib.md5 TypeError in deepcopy (hashlib issue)."""
        d = self.__dict__.copy()
        d.pop('_digest')
        d.pop('_hash_struct')
        return d

    def __setstate__(self, d):
        self._digest = hashlib.md5()
        name, input_count = d['name'], d['input_count']
        self._digest.update(name.encode())
        self._hash_struct = struct.Struct( "!" +" ".join(["I"] * (input_count * 2)))
        self.__dict__.update(d)

    def __call__(self, *vals):
        """
        Given integer or rational inputs, generate a cross-platform,
        architecture-independent 32-bit integer hash.
        """
        # Convert inputs to (numer, denom) pairs with integers
        # becoming (int, 1) pairs to match gmpy2.mpqs for int values.
        pairs = [self._rational(val) for val in vals]
        # Unpack pairs and fill struct with ints to update md5 hash
        ints = [el for pair in pairs for el in pair]
        digest = self._digest.copy()
        digest.update(self._hash_struct.pack(*ints))
        # Convert from hex string to 32 bit int
        return int(digest.hexdigest()[:7], 16)



class TimeAwareRandomState(TimeAware):
    """
    Generic base class to enable time-dependent random
    streams. Although this class is the basis of all random numbergen
    classes, it is designed to be useful whenever time-dependent
    randomness is needed using param's notion of time. For instance,
    this class is used by the imagen package to define time-dependent,
    random distributions over 2D arrays.

    For generality, this class may use either the Random class from
    Python's random module or numpy.random.RandomState. Either of
    these random state objects may be used to generate numbers from
    any of several different random distributions (e.g. uniform,
    Gaussian). The latter offers the ability to generate
    multi-dimensional random arrays and more random distributions but
    requires numpy as a dependency.

    If declared time_dependent, the random state is fully determined
    by a hash value per call. The hash is initialized once with the
    object name and then per call using a tuple consisting of the time
    (via time_fn) and the global param.random_seed.  As a consequence,
    for a given name and fixed value of param.random_seed, the random
    values generated will be a fixed function of time.

    If the object name has not been set and time_dependent is True, a
    message is generated warning that the default object name is
    dependent on the order of instantiation.  To ensure that the
    random number stream will remain constant even if other objects
    are added or reordered in your file, supply a unique name
    explicitly when you construct the RandomDistribution object.
    """

    # Historically, the default random state was seeded with the tuple
    # (500, 500). The CPython implementation implicitly formed an unsigned
    # integer seed using the hash of the tuple as in the expression below. Note
    # that the resulting integer, and therefore the default initial random
    # state, varies across CPython versions (as the hash algorithm has changed)
    # and also between 32-bit and 64-bit interpreters.
    #
    # Seeding based on hashing is deprecated since Python 3.9 and removed in
    # Python 3.11; we explicitly continue the historical behavior for the time
    # being.
    random_generator = param.Parameter(
        default=random.Random(c_size_t(hash((500,500))).value), doc=
        """
        Random state used by the object. This may be an instance
        of random.Random from the Python standard library or an
        instance of numpy.random.RandomState.

        This random state may be exclusively owned by the object or
        may be shared by all instance of the same class. It is always
        possible to give an object its own unique random state by
        setting this parameter with a new random state instance.
        """)

    __abstract = True

    def _initialize_random_state(self, seed=None, shared=True, name=None):
        """
        Initialize the random state correctly.

        Method to be called in the constructor of
        subclasses to initialize the random state correctly.

        If seed is None, there is no control over the random stream
        (no reproducibility of the stream).

        If shared is True (and not time-dependent), the random state
        is shared across all objects of the given class. This can be
        overridden per object by creating new random state to assign
        to the random_generator parameter.
        """
        if seed is None: # Equivalent to an uncontrolled seed.
            seed = random.Random().randint(0, 1000000)
            suffix = ''
        else:
            suffix = str(seed)

        # If time_dependent, independent state required: otherwise
        # time-dependent seeding (via hash) will affect shared
        # state. Note that if all objects have time_dependent=True
        # shared random state is safe and more memory efficient.
        if self.time_dependent or not shared:
            self.random_generator = type(self.random_generator)(seed)

        # Seed appropriately (if not shared)
        if not shared:
            self.random_generator.seed(seed)

        if name is None:
            self._verify_constrained_hash()

        hash_name = name if name else self.name
        if not shared:  hash_name += suffix
        self._hashfn = Hash(hash_name, input_count=2)

        if self.time_dependent:
            self._hash_and_seed()


    def _verify_constrained_hash(self):
        """Warn if the object name is not explicitly set."""
        changed_params = self.param.values(onlychanged=True)
        if self.time_dependent and ('name' not in changed_params):
            self.param.log(param.WARNING, "Default object name used to set the seed: "
                           "random values conditional on object instantiation order.")

    def _hash_and_seed(self):
        """
        To be called between blocks of random number generation. A
        'block' can be an unbounded sequence of random numbers so long
        as the time value (as returned by time_fn) is guaranteed not
        to change within the block. If this condition holds, each
        block of random numbers is time-dependent.

        Note: param.random_seed is assumed to be integer or rational.
        """
        hashval = self._hashfn(self.time_fn(), param.random_seed)
        self.random_generator.seed(hashval)



class RandomDistribution(NumberGenerator, TimeAwareRandomState):
    """
    The base class for all Numbergenerators using random state.

    Numbergen provides a hierarchy of classes to make it easier to use
    the random distributions made available in Python's random module,
    where each class is tied to a particular random distribution.

    RandomDistributions support setting parameters on creation rather
    than passing them each call, and allow pickling to work properly.
    Code that uses these classes will be independent of how many
    parameters are used by the underlying distribution, and can simply
    treat them as a generic source of random numbers.

    RandomDistributions are examples of TimeAwareRandomState, and thus
    can be locked to a global time if desired.  By default,
    time_dependent=False, and so a new random value will be generated
    each time these objects are called.  If you have a global time
    function, you can set time_dependent=True, so that the random
    values will instead be constant at any given time, changing only
    when the time changes.  Using time_dependent values can help you
    obtain fully reproducible streams of random numbers, even if you
    e.g. move time forwards and backwards for testing.

    Note: Each RandomDistribution object has independent random state.
    """

    seed = param.Integer(default=None, allow_None=True, doc="""
       Sets the seed of the random number generator and can be used to
       randomize time dependent streams.

       If seed is None, there is no control over the random stream
       (i.e. no reproducibility of the stream).""")

    __abstract = True

    def __init__(self, **params):
        """
        Initialize a new Random() instance and store the supplied
        positional and keyword arguments.

        If seed=X is specified, sets the Random() instance's seed.
        Otherwise, calls creates an unseeded Random instance which is
        likely to result in a state very different from any just used.
        """
        super().__init__(**params)
        self._initialize_random_state(seed=self.seed, shared=False)

    def __call__(self):
        if self.time_dependent:
            self._hash_and_seed()


class UniformRandom(RandomDistribution):
    """
    Specified with lbound and ubound; when called, return a random
    number in the range [lbound, ubound).

    See the random module for further details.
    """

    lbound = param.Number(default=0.0,doc="Inclusive lower bound.")

    ubound = param.Number(default=1.0,doc="Exclusive upper bound.")


    def __call__(self):
        super().__call__()
        return self.random_generator.uniform(self.lbound,self.ubound)



class UniformRandomOffset(RandomDistribution):
    """
    Identical to UniformRandom, but specified by mean and range.
    When called, return a random number in the range
    [mean - range/2, mean + range/2).

    See the random module for further details.
    """

    mean = param.Number(default=0.0, doc="""Mean value""")

    range = param.Number(default=1.0, bounds=(0.0,None), doc="""
        Difference of maximum and minimum value""")


    def __call__(self):
        super().__call__()
        return self.random_generator.uniform(
                self.mean - self.range / 2.0,
                self.mean + self.range / 2.0)



class UniformRandomInt(RandomDistribution):
    """
    Specified with lbound and ubound; when called, return a random
    number in the inclusive range [lbound, ubound].

    See the randint function in the random module for further details.
    """

    lbound = param.Number(default=0,doc="Inclusive lower bound.")
    ubound = param.Number(default=1000,doc="Inclusive upper bound.")


    def __call__(self):
        super().__call__()
        x = self.random_generator.randint(self.lbound,self.ubound)
        return x



class Choice(RandomDistribution):
    """
    Return a random element from the specified list of choices.

    Accepts items of any type, though they are typically numbers.
    See the choice() function in the random module for further details.
    """

    choices = param.List(default=[0,1],
        doc="List of items from which to select.")


    def __call__(self):
        super().__call__()
        return self.random_generator.choice(self.choices)



class NormalRandom(RandomDistribution):
    """
    Normally distributed (Gaussian) random number.

    Specified with mean mu and standard deviation sigma.
    See the random module for further details.
    """

    mu = param.Number(default=0.0,doc="Mean value.")

    sigma = param.Number(default=1.0,bounds=(0.0,None),doc="Standard deviation.")


    def __call__(self):
        super().__call__()
        return self.random_generator.normalvariate(self.mu,self.sigma)



class VonMisesRandom(RandomDistribution):
    """
    Circularly normal distributed random number.

    If kappa is zero, this distribution reduces to a uniform random
    angle over the range 0 to 2*pi.  Otherwise, it is concentrated to
    a greater or lesser degree (determined by kappa) around the mean
    mu.  For large kappa (narrow peaks), this distribution approaches
    the Gaussian (normal) distribution with variance 1/kappa.  See the
    random module for further details.
    """

    mu = param.Number(default=0.0,softbounds=(0.0,2*pi),doc="""
        Mean value, typically in the range 0 to 2*pi.""")

    kappa = param.Number(default=1.0,bounds=(0.0,None),softbounds=(0.0,50.0),doc="""
        Concentration (inverse variance).""")


    def __call__(self):
        super().__call__()
        return self.random_generator.vonmisesvariate(self.mu,self.kappa)




class ScaledTime(NumberGenerator, TimeDependent):
    """The current time multiplied by some conversion factor."""

    factor = param.Number(default=1.0, doc="""
       The factor to be multiplied by the current time value.""")


    def __call__(self):
        return float(self.time_fn() * self.factor)



class BoxCar(NumberGenerator, TimeDependent):
    """
    The boxcar function over the specified time interval. The bounds
    are exclusive: zero is returned at the onset time and at the
    offset (onset+duration).

    If duration is None, then this reduces to a step function around the
    onset value with no offset.

    See http://en.wikipedia.org/wiki/Boxcar_function
    """

    onset = param.Number(0.0, doc="Time of onset.")

    duration = param.Number(None, allow_None=True, bounds=(0.0,None), doc="""
        Duration of step value.""")


    def __call__(self):
        if self.time_fn() <= self.onset:
            return 0.0
        elif (self.duration is not None) and (self.time_fn() > self.onset + self.duration):
            return 0.0
        else:
            return 1.0



class SquareWave(NumberGenerator, TimeDependent):
    """
    Generate a square wave with 'on' periods returning 1.0 and
    'off'periods returning 0.0 of specified duration(s). By default
    the portion of time spent in the high state matches the time spent
    in the low state (a duty cycle of 50%), but the duty cycle can be
    controlled if desired.

    The 'on' state begins after a time specified by the 'onset'
    parameter.  The onset duration supplied must be less than the off
    duration.
    """

    onset = param.Number(0.0, doc="""Time of onset of the first 'on'
        state relative to time 0. Must be set to a value less than the
        'off_duration' parameter.""")

    duration = param.Number(1.0, allow_None=False, bounds=(0.0,None), doc="""
         Duration of the 'on' state during which a value of 1.0 is
         returned.""")

    off_duration = param.Number(default=None, allow_None=True,
                                bounds=(0.0,None), doc="""
        Duration of the 'off' value state during which a value of 0.0
        is returned. By default, this duration matches the value of
        the 'duration' parameter.""")


    def __init__(self, **params):
        super().__init__(**params)

        if self.off_duration is None:
            self.off_duration = self.duration

        if self.onset > self.off_duration:
            raise AssertionError("Onset value needs to be less than %s" % self.onset)


    def __call__(self):
        phase_offset = (self.time_fn() - self.onset) % (self.duration + self.off_duration)
        if phase_offset < self.duration:
            return 1.0
        else:
            return 0.0



class ExponentialDecay(NumberGenerator, TimeDependent):
    """
    Function object that provides a value that decays according to an
    exponential function, based on a given time function.

    Returns starting_value*base^(-time/time_constant).

    See http://en.wikipedia.org/wiki/Exponential_decay.
    """

    starting_value = param.Number(1.0, doc="Value used for time zero.")

    ending_value = param.Number(0.0, doc="Value used for time infinity.")

    time_constant = param.Number(10000,doc="""
        Time scale for the exponential; large values give slow decay.""")

    base = param.Number(e, doc="""
        Base of the exponent; the default yields starting_value*exp(-t/time_constant).
        Another popular choice of base is 2, which allows the
        time_constant to be interpreted as a half-life.""")


    def __call__(self):
        Vi = self.starting_value
        Vm = self.ending_value
        exp = -1.0*float(self.time_fn())/float(self.time_constant)
        return Vm + (Vi - Vm) * self.base**exp



class TimeSampledFn(NumberGenerator, TimeDependent):
    """
    Samples the values supplied by a time_dependent callable at
    regular intervals of duration 'period', with the sampled value
    held constant within each interval.
    """

    period = param.Number(default=1.0, bounds=(0.0,None),
        inclusive_bounds=(False,True), softbounds=(0.0,5.0), doc="""
        The periodicity with which the values of fn are sampled.""")

    offset = param.Number(default=0.0, bounds=(0.0,None),
                          softbounds=(0.0,5.0), doc="""
        The offset from time 0.0 at which the first sample will be drawn.
        Must be less than the value of period.""")

    fn = param.Callable(doc="""
        The time-dependent function used to generate the sampled values.""")


    def __init__(self, **params):
        super().__init__(**params)

        if not getattr(self.fn,'time_dependent', False):
            raise Exception("The function 'fn' needs to be time dependent.")

        if self.time_fn != self.fn.time_fn:
            raise Exception("Objects do not share the same time_fn")

        if self.offset >= self.period:
            raise Exception("The onset value must be less than the period.")


    def __call__(self):
        current_time = self.time_fn()
        current_time += self.offset
        difference = current_time % self.period
        with self.time_fn as t:
            t(current_time - difference - self.offset)
            value = self.fn()
        return value



class BoundedNumber(NumberGenerator):
    """
    Function object that silently enforces numeric bounds on values
    returned by a callable object.
    """

    generator = param.Callable(None, doc="Object to call to generate values.")

    bounds = param.Parameter((None,None), doc="""
        Legal range for the value returned, as a pair.

        The default bounds are (None,None), meaning there are actually
        no bounds.  One or both bounds can be set by specifying a
        value.  For instance, bounds=(None,10) means there is no lower
        bound, and an upper bound of 10.""")


    def __call__(self):
        val = self.generator()
        min_, max_ = self.bounds
        if   min_ is not None and val < min_: return min_
        elif max_ is not None and val > max_: return max_
        else: return val


_public = {_k for _k,_v in locals().items() if isinstance(_v,type) and issubclass(_v,NumberGenerator)}
__all__ = ["__version__", *_public]



"""
This module is deprecated and will be removed in a future release.

Provide consistent and up-to-date ``__version__`` strings for
Python packages.

See https://github.com/holoviz/autover for more information.
"""

# The Version class is a copy of autover.version.Version v0.2.5,
# except as noted below.
#
# The current version of autover supports a workflow based on tagging
# a git repository, and reports PEP440 compliant version information.
# Previously, the workflow required editing of version numbers in
# source code, and the version was not necessarily PEP440 compliant.
# Version.__new__ is added here to provide the previous Version class
# (OldDeprecatedVersion) if Version is called in the old way.


__author__ = 'Jean-Luc Stevens'

import os
import subprocess
import json
import warnings

def run_cmd(args, cwd=None):
    warnings.warn(
        'param.version.run_cmd has been deprecated and will be removed in a future version.',
        FutureWarning,
        stacklevel=2
    )
    kwargs = {}
    if os.name == 'nt':
        kwargs['creationflags'] = subprocess.CREATE_NO_WINDOW
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            cwd=cwd,
                            **kwargs)
    output, error = (str(s.decode()).strip() for s in proc.communicate())

    # Detects errors as _either_ a non-zero return code _or_ messages
    # printed to stderr, because the return code is erroneously fixed at
    # zero in some cases (see https://github.com/holoviz/param/pull/389).
    if proc.returncode != 0 or len(error) > 0:
        raise Exception(proc.returncode, error)
    return output



class Version:
    """
    A simple approach to Python package versioning that supports PyPI
    releases and additional information when working with version
    control. When obtaining a package from PyPI, the version returned
    is a string-formatted rendering of the supplied release tuple.
    For instance, release (1,0) tagged as ``v1.0`` in the version
    control system will return ``1.0`` for ``str(__version__)``.  Any
    number of items can be supplied in the release tuple, with either
    two or three numeric versioning levels typical.

    During development, a command like ``git describe`` will be used to
    compute the number of commits since the last version tag, the short
    commit hash, and whether the commit is dirty (has changes not yet
    committed). Version tags must start with a lowercase 'v' and have a
    period in them, e.g. v2.0, v0.9.8 or v0.1 and may include the PEP440
    prerelease identifiers of 'a' (alpha) 'b' (beta) or 'rc' (release
    candidate) allowing tags such as v2.0.a3, v0.9.8.b3 or v0.1.rc5.

    Also note that when version control system (VCS) information is
    used, the number of commits since the last version tag is
    determined. This approach is often useful in practice to decide
    which version is newer for a single developer, but will not
    necessarily be reliable when comparing against a different fork or
    branch in a distributed VCS.

    For git, if you want version control information available even in
    an exported archive (e.g. a .zip file from GitHub), you can set
    the following line in the .gitattributes file of your project::

      __init__.py export-subst

    Note that to support pip installation directly from GitHub via git
    archive, a .version file must be tracked by the repo to supply the
    release number (otherwise only the short SHA is available).

    The PEP440 format returned is [N!]N(.N)*[{a|b|rc}N][.postN+SHA]
    where everything before .postN is obtained from the tag, the N in
    .postN is the number of commits since the last tag and the SHA is
    obtained via git describe. This later portion is only shown if the
    commit count since the last tag is non zero. Instead of '.post', an
    alternate valid prefix such as '.rev', '_rev', '_r' or '.r' may be
    supplied.
    """

    def __new__(cls,**kw):
        # If called in the old way, provide the previous class. Means
        # PEP440/tag based workflow warning below will never appear.
        if ('release' in kw and kw['release'] is not None) or \
           ('dev' in kw and kw['dev'] is not None) or \
           ('commit_count' in kw):
            return OldDeprecatedVersion(**kw)
        else:
            return super().__new__(cls)


    def __init__(self, release=None, fpath=None, commit=None, reponame=None,
                 commit_count_prefix='.post', archive_commit=None, **kwargs):
        """
        :release:      Release tuple (corresponding to the current VCS tag)
        :commit        Short SHA. Set to '$Format:%h$' for git archive support.
        :fpath:        Set to ``__file__`` to access version control information
        :reponame:     Used to verify VCS repository name.
        """
        warnings.warn(
            'param.version.Version has been deprecated and will be removed in a future version.',
            FutureWarning,
            stacklevel=2
        )
        self.fpath = fpath
        self._expected_commit = commit

        if release is not None or 'commit_count' in kwargs:
            print('WARNING: param.Version now supports PEP440 and a new tag based workflow. See param/version.py for more details')

        self.expected_release = release

        self._commit = None if (commit is None or commit.startswith("$Format")) else commit
        self._commit_count = None
        self._release = None
        self._dirty = False
        self._prerelease = None

        self.archive_commit= archive_commit

        self.reponame = reponame
        self.commit_count_prefix = commit_count_prefix

    @property
    def prerelease(self):
        """
        Either None or one of 'aN' (alpha), 'bN' (beta) or 'rcN'
        (release candidate) where N is an integer.
        """
        return self.fetch()._prerelease

    @property
    def release(self):
        """Return the release tuple."""
        return self.fetch()._release

    @property
    def commit(self):
        """A specification for this particular VCS version, e.g. a short git SHA."""
        return self.fetch()._commit

    @property
    def commit_count(self):
        """Return the number of commits since the last release."""
        return self.fetch()._commit_count

    @property
    def dirty(self):
        """True if there are uncommited changes, False otherwise."""
        return self.fetch()._dirty


    def fetch(self):
        """
        Return a tuple of the major version together with the
        appropriate SHA and dirty bit (for development version only).
        """
        if self._release is not None:
            return self

        self._release = self.expected_release
        if not self.fpath:
            self._commit = self._expected_commit
            return self

         # Only git right now but easily extended to SVN, Mercurial, etc.
        for cmd in ['git', 'git.cmd', 'git.exe']:
            try:
                self.git_fetch(cmd)
                break
            except OSError:
                pass
        return self


    def git_fetch(self, cmd='git', as_string=False):
        commit_argument = self._commit
        output = None
        try:
            if self.reponame is not None:
                # Verify this is the correct repository (since fpath could
                # be an unrelated git repository, and autover could just have
                # been copied/installed into it).
                remotes = run_cmd([cmd, 'remote', '-v'],
                                  cwd=os.path.dirname(self.fpath))
                repo_matches = ['/' + self.reponame + '.git' ,
                                # A remote 'server:reponame.git' can also be referred
                                # to (i.e. cloned) as `server:reponame`.
                                '/' + self.reponame + ' ']
                if not any(m in remotes for m in repo_matches):
                    try:
                        output = self._output_from_file()
                        if output is not None:
                            self._update_from_vcs(output)
                    except Exception: pass
            if output is None:
                # glob pattern (not regexp) matching vX.Y.Z* tags
                output = run_cmd([cmd, 'describe', '--long', '--match',
                                  "v[0-9]*.[0-9]*.[0-9]*", '--dirty'],
                                 cwd=os.path.dirname(self.fpath))
            if as_string: return output
        except Exception as e1:
            try:
                output = self._output_from_file()
                if output is not None:
                    self._update_from_vcs(output)
                if self._known_stale():
                    self._commit_count = None
                if as_string: return output

                # If an explicit commit was supplied (e.g from git
                # archive), it should take precedence over the file.
                if commit_argument:
                    self._commit = commit_argument
                return

            except OSError:
                if e1.args[1] == 'fatal: No names found, cannot describe anything.':
                    raise Exception("Cannot find any git version tags of format v*.*")
                # If there is any other error, return (release value still useful)
                return self

        self._update_from_vcs(output)


    def _known_stale(self):
        """
        Return whether the commit is know to be stale or not.

        The commit is known to be from a file (and therefore stale) if a
        SHA is supplied by git archive and doesn't match the parsed commit.
        """
        if self._output_from_file() is None:
            commit = None
        else:
            commit = self.commit

        known_stale = (self.archive_commit is not None
                       and not self.archive_commit.startswith('$Format')
                       and self.archive_commit != commit)
        if known_stale: self._commit_count = None
        return known_stale

    def _output_from_file(self, entry='git_describe'):
        """
        Read the version from a .version file that may exist alongside __init__.py.

        This file can be generated by piping the following output to file:

        git describe --long --match v*.*
        """
        try:
            vfile = os.path.join(os.path.dirname(self.fpath), '.version')
            with open(vfile) as f:
                return json.loads(f.read()).get(entry, None)
        except Exception: # File may be missing if using pip + git archive
            return None


    def _update_from_vcs(self, output):
        """Update state based on the VCS state e.g the output of git describe."""
        split = output[1:].split('-')
        dot_split = split[0].split('.')
        for prefix in ['a','b','rc']:
            if prefix in dot_split[-1]:
                prefix_split = dot_split[-1].split(prefix)
                self._prerelease = prefix + prefix_split[-1]
                dot_split[-1] = prefix_split[0]


        self._release = tuple(int(el) for el in dot_split)
        self._commit_count = int(split[1])

        self._commit = str(split[2][1:]) # Strip out 'g' prefix ('g'=>'git')

        self._dirty = (split[-1]=='dirty')
        return self

    def __str__(self):
        """
        Version in x.y.z string format. Does not include the "v"
        prefix of the VCS version tags, for pip compatibility.

        If the commit count is non-zero or the repository is dirty,
        the string representation is equivalent to the output of::

          git describe --long --match v*.* --dirty

        (with "v" prefix removed).
        """
        known_stale = self._known_stale()
        if self.release is None and not known_stale:
            extracted_directory_tag = self._output_from_file(entry='extracted_directory_tag')
            return 'None' if extracted_directory_tag is None else extracted_directory_tag
        elif self.release is None and known_stale:
            extracted_directory_tag = self._output_from_file(entry='extracted_directory_tag')
            if extracted_directory_tag is not None:
                return extracted_directory_tag
            return '0.0.0+g{SHA}-gitarchive'.format(SHA=self.archive_commit)

        release = '.'.join(str(el) for el in self.release)
        prerelease = '' if self.prerelease is None else self.prerelease

        if self.commit_count == 0 and not self.dirty:
            return release + prerelease

        commit = self.commit
        dirty = '-dirty' if self.dirty else ''
        archive_commit = ''
        if known_stale:
            archive_commit = '-gitarchive'
            commit = self.archive_commit

        if archive_commit != '':
            postcount = self.commit_count_prefix + '0'
        elif self.commit_count not in [0, None]:
            postcount = self.commit_count_prefix + str(self.commit_count)
        else:
            postcount = ''

        components = [release, prerelease, postcount,
                      '' if commit is None else '+g' + commit, dirty,
                      archive_commit]
        return ''.join(components)

    def __repr__(self):
        return str(self)

    def abbrev(self):
        """Abbreviated string representation of just the release number."""
        return '.'.join(str(el) for el in self.release)

    def verify(self, string_version=None):
        """
        Check that the version information is consistent with the VCS
        before doing a release. If supplied with a string version,
        this is also checked against the current version. Should be
        called from setup.py with the declared package version before
        releasing to PyPI.
        """
        if string_version and string_version != str(self):
            raise Exception("Supplied string version does not match current version.")

        if self.dirty:
            raise Exception("Current working directory is dirty.")

        if self.expected_release is not None and self.release != self.expected_release:
            raise Exception("Declared release does not match current release tag.")

        if self.commit_count !=0:
            raise Exception("Please update the VCS version tag before release.")

        if (self._expected_commit is not None
            and not self._expected_commit.startswith( "$Format")):
            raise Exception("Declared release does not match the VCS version tag")



    @classmethod
    def get_setup_version(cls, setup_path, reponame, describe=False,
                          dirty='report', pkgname=None, archive_commit=None):
        """
        Get the version from the .version file (if available)
        or more up-to-date information from git describe (if available).

        Assumes the __init__.py will be found in the directory
        {reponame}/__init__.py relative to setup.py unless pkgname is
        explicitly specified in which case that name is used instead.

        If describe is True, the raw string obtained from git described is
        returned which is useful for updating the .version file.

        The dirty policy can be one of 'report', 'strip', 'raise'. If it is
        'report' the version string may end in '-dirty' if the repository is
        in a dirty state.  If the policy is 'strip', the '-dirty' suffix
        will be stripped out if present. If the policy is 'raise', an
        exception is raised if the repository is in a dirty state. This can
        be useful if you want to make sure packages are not built from a
        dirty repository state.
        """
        pkgname = reponame if pkgname is None else pkgname
        policies = ['raise','report', 'strip']
        if dirty not in policies:
            raise AssertionError("get_setup_version dirty policy must be in %r" % policies)

        fpath = os.path.join(setup_path, pkgname, "__init__.py")
        version = Version(fpath=fpath, reponame=reponame, archive_commit=archive_commit)
        if describe:
            vstring = version.git_fetch(as_string=True)
        else:
            vstring = str(version)

        if version.dirty and dirty == 'raise':
            raise AssertionError('Repository is in a dirty state.')
        elif version.dirty and dirty=='strip':
            return vstring.replace('-dirty', '')
        else:
            return vstring


    @classmethod
    def extract_directory_tag(cls, setup_path, reponame):
        setup_dir = os.path.split(setup_path)[-1] # Directory containing setup.py
        prefix = reponame + '-' # Prefix to match
        if setup_dir.startswith(prefix):
            tag = setup_dir[len(prefix):]
            # Assuming the tag is a version if it isn't empty, 'master' or 'main' and has a dot in it
            if tag not in ['', 'master', 'main'] and ('.' in tag):
                return tag
        return None


    @classmethod
    def setup_version(cls, setup_path, reponame, archive_commit=None,
                      pkgname=None, dirty='report'):
        info = {}
        git_describe = None
        pkgname = reponame if pkgname is None else pkgname
        try:
            # Will only work if in a git repo and git is available
            git_describe  = Version.get_setup_version(setup_path,
                                                      reponame,
                                                      describe=True,
                                                      dirty=dirty,
                                                      pkgname=pkgname,
                                                      archive_commit=archive_commit)

            if git_describe is not None:
                info['git_describe'] = git_describe
        except Exception: pass

        if git_describe is None:
            extracted_directory_tag = Version.extract_directory_tag(setup_path, reponame)
            if extracted_directory_tag is not None:
                info['extracted_directory_tag'] = extracted_directory_tag
            try:
                with open(os.path.join(setup_path, pkgname, '.version'), 'w') as f:
                    f.write(json.dumps({'extracted_directory_tag':extracted_directory_tag}))
            except Exception:
                print('Error in setup_version: could not write .version file.')


        info['version_string'] =  Version.get_setup_version(setup_path,
                                                            reponame,
                                                            describe=False,
                                                            dirty=dirty,
                                                            pkgname=pkgname,
                                                            archive_commit=archive_commit)
        try:
            with open(os.path.join(setup_path, pkgname, '.version'), 'w') as f:
                f.write(json.dumps(info))
        except Exception:
            print('Error in setup_version: could not write .version file.')

        return info['version_string']



def get_setup_version(location, reponame, pkgname=None, archive_commit=None):
    """
    Get the current version from either git describe or the
    .version file (if available).

    Set pkgname to the package name if it is different from the
    repository name.

    To ensure git information is included in a git archive, add
    setup.py to .gitattributes (in addition to __init__):
    ```
    __init__.py export-subst
    setup.py export-subst
    ```
    Then supply "$Format:%h$" for archive_commit.

    """
    import warnings
    warnings.warn(
        'param.version.get_setup_version has been deprecated and will be removed in a future version.',
        FutureWarning,
        stacklevel=2
    )
    pkgname = reponame if pkgname is None else pkgname
    if archive_commit is None:
        warnings.warn("No archive commit available; git archives will not contain version information")
    return Version.setup_version(os.path.dirname(os.path.abspath(location)),reponame,pkgname=pkgname,archive_commit=archive_commit)


def get_setupcfg_version():
    """
    As get_setup_version(), but configure via setup.cfg.

    If your project uses setup.cfg to configure setuptools, and hence has
    at least a "name" key in the [metadata] section, you can
    set the version as follows:
    ```
    [metadata]
    name = mypackage
    version = attr: autover.version.get_setup_version2
    ```

    If the repository name is different from the package name, specify
    `reponame` as a [tool:autover] option:
    ```
    [tool:autover]
    reponame = mypackage
    ```

    To ensure git information is included in a git archive, add
    setup.cfg to .gitattributes (in addition to __init__):
    ```
    __init__.py export-subst
    setup.cfg export-subst
    ```

    Then add the following to setup.cfg:
    ```
    [tool:autover.configparser_workaround.archive_commit=$Format:%h$]
    ```

    The above being a section heading rather than just a key is
    because setuptools requires % to be escaped with %, or it can't
    parse setup.cfg...but then git export-subst would not work.

    """
    import configparser
    import re
    warnings.warn(
        'param.version.get_setupcfg_version has been deprecated and will be removed in a future version.',
        FutureWarning,
        stacklevel=2
    )
    cfg = "setup.cfg"
    autover_section = 'tool:autover'
    config = configparser.ConfigParser()
    config.read(cfg)
    pkgname = config.get('metadata','name')
    reponame = config.get(autover_section,'reponame',vars={'reponame':pkgname}) if autover_section in config.sections() else pkgname

    ###
    # hack archive_commit into section heading; see docstring
    archive_commit = None
    archive_commit_key = autover_section+'.configparser_workaround.archive_commit'
    for section in config.sections():
        if section.startswith(archive_commit_key):
            archive_commit = re.match(r".*=\s*(\S*)\s*",section).group(1)
    ###
    return get_setup_version(cfg,reponame=reponame,pkgname=pkgname,archive_commit=archive_commit)


# from param/version.py aa087db29976d9b7e0f59c29789dfd721c85afd0
class OldDeprecatedVersion:
    """
    A simple approach to Python package versioning that supports PyPI
    releases and additional information when working with version
    control. When obtaining a package from PyPI, the version returned
    is a string-formatted rendering of the supplied release tuple.
    For instance, release (1,0) tagged as ``v1.0`` in the version
    control system will return ``1.0`` for ``str(__version__)``.  Any
    number of items can be supplied in the release tuple, with either
    two or three numeric versioning levels typical.

    During development, a command like ``git describe`` will be used to
    compute the number of commits since the last version tag, the
    short commit hash, and whether the commit is dirty (has changes
    not yet committed). Version tags must start with a lowercase 'v'
    and have a period in them, e.g. v2.0, v0.9.8 or v0.1.

    Development versions are supported by setting the dev argument to an
    appropriate dev version number. The corresponding tag can be PEP440
    compliant (using .devX) of the form v0.1.dev3, v1.9.0.dev2 etc but
    it doesn't have to be as the dot may be omitted i.e v0.1dev3,
    v1.9.0dev2 etc.

    Also note that when version control system (VCS) information is
    used, the comparison operators take into account the number of
    commits since the last version tag. This approach is often useful
    in practice to decide which version is newer for a single
    developer, but will not necessarily be reliable when comparing
    against a different fork or branch in a distributed VCS.

    For git, if you want version control information available even in
    an exported archive (e.g. a .zip file from GitHub), you can set
    the following line in the .gitattributes file of your project::

      __init__.py export-subst
    """

    def __init__(self, release=None, fpath=None, commit=None,
                 reponame=None, dev=None, commit_count=0):
        """
        :release:      Release tuple (corresponding to the current VCS tag)
        :commit        Short SHA. Set to '$Format:%h$' for git archive support.
        :fpath:        Set to ``__file__`` to access version control information
        :reponame:     Used to verify VCS repository name.
        :dev:          Development version number. None if not a development version.
        :commit_count  Commits since last release. Set for dev releases.
        """
        warnings.warn(
            'param.version.OldDeprecatedVersion has been deprecated and will be removed in a future version.',
            FutureWarning,
            stacklevel=2
        )
        self.fpath = fpath
        self._expected_commit = commit
        self.expected_release = release

        self._commit = None if commit in [None, "$Format:%h$"] else commit
        self._commit_count = commit_count
        self._release = None
        self._dirty = False
        self.reponame = reponame
        self.dev = dev

    @property
    def release(self):
        """Return the release tuple."""
        return self.fetch()._release

    @property
    def commit(self):
        """A specification for this particular VCS version, e.g. a short git SHA."""
        return self.fetch()._commit

    @property
    def commit_count(self):
        """Return the number of commits since the last release."""
        return self.fetch()._commit_count

    @property
    def dirty(self):
        """True if there are uncommited changes, False otherwise."""
        return self.fetch()._dirty


    def fetch(self):
        """
        Return a tuple of the major version together with the
        appropriate SHA and dirty bit (for development version only).
        """
        if self._release is not None:
            return self

        self._release = self.expected_release
        if not self.fpath:
            self._commit = self._expected_commit
            return self

         # Only git right now but easily extended to SVN, Mercurial, etc.
        for cmd in ['git', 'git.cmd', 'git.exe']:
            try:
                self.git_fetch(cmd)
                break
            except OSError:
                pass
        return self


    def git_fetch(self, cmd='git'):
        try:
            if self.reponame is not None:
                # Verify this is the correct repository (since fpath could
                # be an unrelated git repository, and param could just have
                # been copied/installed into it).
                output = run_cmd([cmd, 'remote', '-v'],
                                 cwd=os.path.dirname(self.fpath))
                repo_matches = ['/' + self.reponame + '.git' ,
                                # A remote 'server:reponame.git' can also be referred
                                # to (i.e. cloned) as `server:reponame`.
                                '/' + self.reponame + ' ']
                if not any(m in output for m in repo_matches):
                    return self

            output = run_cmd([cmd, 'describe', '--long', '--match', 'v*.*', '--dirty'],
                             cwd=os.path.dirname(self.fpath))
        except Exception as e:
            if e.args[1] == 'fatal: No names found, cannot describe anything.':
                raise Exception("Cannot find any git version tags of format v*.*")
            # If there is any other error, return (release value still useful)
            return self

        self._update_from_vcs(output)

    def _update_from_vcs(self, output):
        """Update state based on the VCS state e.g the output of git describe."""
        split = output[1:].split('-')
        if 'dev' in split[0]:
            dev_split = split[0].split('dev')
            self.dev = int(dev_split[1])
            split[0] = dev_split[0]
            # Remove the pep440 dot if present
            if split[0].endswith('.'):
                split[0] = dev_split[0][:-1]

        self._release = tuple(int(el) for el in split[0].split('.'))
        self._commit_count = int(split[1])
        self._commit = str(split[2][1:]) # Strip out 'g' prefix ('g'=>'git')
        self._dirty = (split[-1]=='dirty')
        return self


    def __str__(self):
        """
        Version in x.y.z string format. Does not include the "v"
        prefix of the VCS version tags, for pip compatibility.

        If the commit count is non-zero or the repository is dirty,
        the string representation is equivalent to the output of::

          git describe --long --match v*.* --dirty

        (with "v" prefix removed).
        """
        if self.release is None: return 'None'
        release = '.'.join(str(el) for el in self.release)
        release = '%s.dev%d' % (release, self.dev) if self.dev is not None else release

        if (self._expected_commit is not None) and  ("$Format" not in self._expected_commit):
            pass  # Concrete commit supplied - print full version string
        elif (self.commit_count == 0 and not self.dirty):
            return release

        dirty_status = '-dirty' if self.dirty else ''
        return '{}-{}-g{}{}'.format(release, self.commit_count if self.commit_count else 'x',
                                self.commit, dirty_status)

    def __repr__(self):
        return str(self)

    def abbrev(self,dev_suffix=""):
        """
        Abbreviated string representation, optionally declaring whether it is
        a development version.
        """
        return '.'.join(str(el) for el in self.release) + \
            (dev_suffix if self.commit_count > 0 or self.dirty else "")


    def __eq__(self, other):
        """
        Two versions are considered equivalent if and only if they are
        from the same release, with the same commit count, and are not
        dirty.  Any dirty version is considered different from any
        other version, since it could potentially have any arbitrary
        changes even for the same release and commit count.
        """
        if self.dirty or other.dirty: return False
        return ((self.release, self.commit_count, self.dev)
                == (other.release, other.commit_count, other.dev))

    def __gt__(self, other):
        if self.release == other.release:
            if self.dev == other.dev:
                return self.commit_count > other.commit_count
            elif None in [self.dev, other.dev]:
                return self.dev is None
            else:
                return self.dev > other.dev
        else:
            return (self.release, self.commit_count) > (other.release, other.commit_count)

    def __lt__(self, other):
        if self==other:
            return False
        else:
            return not (self > other)


    def verify(self, string_version=None):
        """
        Check that the version information is consistent with the VCS
        before doing a release. If supplied with a string version,
        this is also checked against the current version. Should be
        called from setup.py with the declared package version before
        releasing to PyPI.
        """
        if string_version and string_version != str(self):
            raise Exception("Supplied string version does not match current version.")

        if self.dirty:
            raise Exception("Current working directory is dirty.")

        if self.release != self.expected_release:
            raise Exception("Declared release does not match current release tag.")

        if self.commit_count !=0:
            raise Exception("Please update the VCS version tag before release.")

        if self._expected_commit not in [None, "$Format:%h$"]:
            raise Exception("Declared release does not match the VCS version tag")




"""
Param: A declarative framework for managing parameters and reactive programming in Python.

Param is a lightweight library for defining and managing user-modifiable parameters,
designed to simplify Python programs and enhance their readability, maintainability,
and robustness. In addition Param provides the `rx` framework for reactive programming.

Param is well-suited for use in scientific computing, data analysis tools,
graphical user interfaces (GUIs), and any Python application where well-defined,
validated parameters are needed.

References
----------
For detailed documentation, see https://param.holoviz.org/.

Examples
--------
Here is an example of using `param.Parameterized` to define a class with validated parameters:

>>> import param
>>> class MyClass(param.Parameterized):
...     my_number = param.Number(default=1, bounds=(0, 10))
...     my_list = param.List(default=[1, 2, 3], item_type=int)

>>> obj = MyClass()
>>> obj.my_number = 5  # Valid
>>> obj.my_number = 15  # Raises ValueError: must be in range (0, 10)

Here is an example of using `param.rx` to define a reactive expression:

>>> import param
>>> rx_value = param.rx([1,2,3])
>>> rx_value.rx.len()
3

Lets update the reactive value and check its length:

>>> rx_value.rx.value = [1,2,3,4]
>>> rx_value.rx.len()
4
"""
import os

from . import version
from .depends import depends
from .parameterized import (
    Parameterized, Parameter, Skip, String, ParameterizedFunction,
    ParamOverrides, Undefined, get_logger, ParameterizedABC,
)
from .parameterized import (output, script_repr,
                            discard_events, edit_constant)
from .parameterized import shared_parameters
from .parameterized import logging_level
from .parameterized import DEBUG, VERBOSE, INFO, WARNING, ERROR, CRITICAL
from .parameters import (
    guess_param_types,
    param_union,
    parameterized_class,
    guess_bounds,
    get_soft_bounds,
    resolve_path,
    Time,
    Infinity,
    Dynamic,
    Bytes,
    Number,
    Integer,
    Magnitude,
    Boolean,
    Tuple,
    NumericTuple,
    XYCoordinates,
    Callable,
    Action,
    Composite,
    SelectorBase,
    ListProxy,
    Selector,
    ObjectSelector,
    ClassSelector,
    List,
    HookList,
    Dict,
    Array,
    DataFrame,
    Series,
    Path,
    Filename,
    Foldername,
    FileSelector,
    ListSelector,
    MultiFileSelector,
    Date,
    CalendarDate,
    Color,
    Range,
    DateRange,
    CalendarDateRange,
    Event,
)
from .reactive import bind, rx
from ._utils import (
    descendents,
    concrete_descendents,
    exceptions_summarized,
    _is_number,
)


# Define '__version__'
try:
    # For performance reasons on imports, avoid importing setuptools_scm
    # if not in a .git folder
    if os.path.exists(os.path.join(os.path.dirname(__file__), "..", ".git")):
        # If setuptools_scm is installed (e.g. in a development environment with
        # an editable install), then use it to determine the version dynamically.
        from setuptools_scm import get_version

        # This will fail with LookupError if the package is not installed in
        # editable mode or if Git is not installed.
        __version__ = get_version(root="..", relative_to=__file__)
    else:
        raise FileNotFoundError
except (ImportError, LookupError, FileNotFoundError):
    # As a fallback, use the version that is hard-coded in the file.
    try:
        # __version__ was added in _version in setuptools-scm 7.0.0, we rely on
        # the hopefully stable version variable.
        from ._version import version as __version__
    except (ModuleNotFoundError, ImportError):
        # Either _version doesn't exist (ModuleNotFoundError) or version isn't
        # in _version (ImportError). ModuleNotFoundError is a subclass of
        # ImportError, let's be explicit anyway.

        # Try something else:
        from importlib.metadata import version as mversion, PackageNotFoundError

        try:
            __version__ = mversion("param")
        except PackageNotFoundError:
            # The user is probably trying to run this without having installed
            # the package.
            __version__ = "0.0.0+unknown"

#: Top-level object to allow messaging not tied to a particular
#: Parameterized object, as in 'param.main.warning("Invalid option")'.
main=Parameterized(name="main")


# A global random seed (integer or rational) available for controlling
# the behaviour of Parameterized objects with random state.
random_seed = 42

__all__ = (
    'Action',
    'Array',
    'Boolean',
    'Bytes',
    'CRITICAL',
    'CalendarDate',
    'CalendarDateRange',
    'Callable',
    'ClassSelector',
    'Color',
    'Composite',
    'DEBUG',
    'DataFrame',
    'Date',
    'DateRange',
    'Dict',
    'Dynamic',
    'ERROR',
    'Event',
    'FileSelector',
    'Filename',
    'Foldername',
    'HookList',
    'INFO',
    'Infinity',
    'Integer',
    'List',
    'ListProxy',
    'ListSelector',
    'Magnitude',
    'MultiFileSelector',
    'Number',
    'NumericTuple',
    'ObjectSelector',
    'ParamOverrides',
    'Parameter',
    'Parameterized',
    'ParameterizedABC',
    'ParameterizedFunction',
    'Path',
    'Range',
    'Selector',
    'SelectorBase',
    'Series',
    'Skip',
    'String',
    'Time',
    'Tuple',
    'Undefined',
    'VERBOSE',
    'WARNING',
    'XYCoordinates',
    '__version__',
    '_is_number',
    'bind',
    'concrete_descendents',
    'depends',
    'descendents',
    'discard_events',
    'edit_constant',
    'exceptions_summarized',
    'get_logger',
    'get_soft_bounds',
    'guess_bounds',
    'guess_param_types',
    'logging_level',
    'main',
    'output',
    'param_union',
    'parameterized_class',
    'random_seed',
    'resolve_path',
    'rx',
    'script_repr',
    'serializer',
    'shared_parameters',
    'version',
)



"""
Reactive API for Dynamic Expression Pipelines.

`rx` provides a wrapper around Python objects, enabling the creation of
reactive expression pipelines that dynamically update based on changes to their
underlying parameters or widgets. This approach simplifies the implementation of
dynamic, interactive behavior in Python code, particularly for data manipulation
and visualization tasks.

Overview
--------

An `rx` instance tracks and records the operations applied to a wrapped object,
constructing a chain of transformations that can reactively update as inputs change.
These operations are recorded and evaluated lazily, ensuring that the object state
remains up-to-date while enabling efficient re-computation.

The original input object is stored in a shared mutable list accessible via the
`_obj` property. This shared reference allows all `rx` instances derived from the
same object to reflect updates, whether through the `.value` property or due to
modifications in the underlying data.

Core Features
-------------

The `rx` implementation intercepts and records various operations performed on an
object, including:

- **Attribute Access**: Captured using `__getattribute__` to watch for method or
  property usage.
- **Method Calls**: Handled through `__call__` after an attribute is accessed.
- **Indexing**: Tracked via `__getitem__` for slicing or item access.
- **Operators**: Supports all standard operators (e.g., `__add__`, `__gt__`) to
  enable arithmetic and logical expressions.
- **NumPy Integration**: Overrides `__array_ufunc__` to capture NumPy universal
  function calls.

This functionality makes `rx` objects particularly well-suited for interactive
environments such as Jupyter notebooks or Python REPLs. Lazy evaluation ensures that
pipeline operations are executed only when their result is needed, enabling dynamic
inspection and interactive exploration. For instance, tab-completion and docstring
retrieval (`dfi.A.max?`) work seamlessly with reactive expressions.

Reactive Expression Mechanics
-----------------------------

When an operation is applied to an `rx` instance:

1. The operation is recorded in a structured format (stored in `_operation`).
2. A new `rx` instance is created using the `_clone` method.
3. The `_dirty` attribute is set to `True` for dependent objects, indicating the
   need for re-evaluation.

Each `_operation` dictionary includes:

- `fn`: The function or method to execute.
- `args`: Positional arguments for the function or method.
- `kwargs`: Keyword arguments for the function or method.
- `reverse`: A flag indicating whether the input object and first argument should
  be swapped during execution.

Chain Tracking and Dependencies
-------------------------------

Reactive pipelines consist of chains of `rx` instances, with each instance linked
to its predecessor. The `_depth` attribute indicates the chain depth, starting at
0 for the root object. Note that `_depth` reflects the chain's outer structure, not
the total number of reactive instances in a pipeline.

Instances also track their dependencies to ensure accurate updates:
- `_method`: Temporarily stores the method or attribute accessed (e.g., `'head'`
  in `dfi.head()`).
- `_dirty`: Indicates whether the current value needs re-computation.
- `_current`: Stores the result of the most recent computation.

Benefits and Use Cases
----------------------

The reactive approach eliminates the need for explicit callbacks and manual state
management, making it ideal for:

- Real-time data processing and visualization pipelines.
- Interactive dashboards and applications.
- Simplifying complex workflows with dynamic dependencies.

By leveraging Python's operator overloading and lazy evaluation, `rx` provides a
powerful and intuitive way to manage dynamic behavior in Python applications.
"""
from __future__ import annotations

import inspect
import math
import operator

from collections.abc import Iterable, Iterator
from functools import partial
from types import FunctionType, MethodType
from typing import Any, Callable, Optional

from .depends import depends
from .display import _display_accessors, _reactive_display_objs
from .parameterized import (
    Parameter, Parameterized, Skip, Undefined, eval_function_with_deps, get_method_owner,
    register_reference_transform, resolve_ref, resolve_value, transform_reference
)
from .parameters import Boolean, Event
from ._utils import _to_async_gen, iscoroutinefunction, full_groupby


class Wrapper(Parameterized):
    """Helper class to allow updating literal values easily."""

    object = Parameter(allow_refs=False)


class GenWrapper(Parameterized):
    """Helper class to allow streaming from generator functions."""

    object = Parameter(allow_refs=True)


class Trigger(Parameterized):
    """Helper class to allow triggering an event under some condition."""

    value = Event()

    def __init__(self, parameters=None, internal=False, **params):
        super().__init__(**params)
        self.internal = internal
        self.parameters = parameters

class Resolver(Parameterized):
    """Helper class to allow (recursively) resolving references."""

    object = Parameter(allow_refs=True)

    recursive = Boolean(default=False)

    value = Parameter()

    def __init__(self, **params):
        self._watchers = []
        super().__init__(**params)

    def _resolve_value(self, *events):
        nested = self.param.object.nested_refs
        refs = resolve_ref(self.object, nested)
        value = resolve_value(self.object, nested)
        if self.recursive:
            new_refs = [r for r in resolve_ref(value, nested) if r not in refs]
            while new_refs:
                refs += new_refs
                value = resolve_value(value, nested)
                new_refs = [r for r in resolve_ref(value, nested) if r not in refs]
            if events:
                self._update_refs(refs)
        self.value = value
        return refs

    @depends('object', watch=True, on_init=True)
    def _resolve_object(self):
        refs = self._resolve_value()
        self._update_refs(refs)

    def _update_refs(self, refs):
        for w in self._watchers:
            (w.inst or w.cls).param.unwatch(w)
        self._watchers = []
        for _, params in full_groupby(refs, lambda x: id(x.owner)):
            self._watchers.append(
                params[0].owner.param.watch(self._resolve_value, [p.name for p in params])
            )


class NestedResolver(Resolver):

    object = Parameter(allow_refs=True, nested_refs=True)


class reactive_ops:
    """
    The reactive operations namespace.

    Provides reactive versions of operations that cannot be made reactive through
    operator overloading. This includes operations such as ``.rx.and_`` and ``.rx.bool``.

    Calling this namespace (``()``) creates and returns a reactive expression, enabling
    dynamic updates and computation tracking.

    Returns
    -------
    rx
        A reactive expression representing the operation applied to the current value.

    References
    ----------
    For more details, see the user guide:
    https://param.holoviz.org/user_guide/Reactive_Expressions.html#special-methods-on-rx

    Examples
    --------
    Create a Parameterized instance and access its reactive operations property:

    >>> import param
    >>> class P(param.Parameterized):
    ...     a = param.Number()
    >>> p = P(a=1)

    Retrieve the current value reactively:

    >>> a_value = p.param.a.rx.value

    Create a reactive expression by calling the namespace:

    >>> rx_expression = p.param.a.rx()

    Use special methods from the reactive ops namespace for reactive operations:

    >>> condition = p.param.a.rx.and_(True)
    >>> piped = p.param.a.rx.pipe(lambda x: x * 2)

    """

    def __init__(self, reactive):
        self._reactive = reactive

    def _as_rx(self):
        return self._reactive if isinstance(self._reactive, rx) else self()

    def __call__(self) -> 'rx':
        """Create a reactive expression."""
        rxi = self._reactive
        return rxi if isinstance(rxi, rx) else rx(rxi)

    def and_(self, other) -> 'rx':
        """
        Perform a logical AND operation with the given operand.

        This method computes a logical AND (``and``) operation between the current
        value of the reactive expression and the provided operand. The result is
        returned as a new reactive expression.

        Parameters
        ----------
        other : any
            The operand to combine with the current value using the AND operation.

        Returns
        -------
        rx
            A new reactive expression representing the result of the AND operation.

        Examples
        --------
        Create two reactive expressions and combine them using ``and_``:

        >>> import param
        >>> a = param.rx(True)
        >>> b = param.rx(False)
        >>> result = a.rx.and_(b)
        >>> result.rx.value
        False

        Combine a reactive expression with a static value:

        >>> result = a.rx.and_(True)
        >>> result.rx.value
        True
        """
        return self._as_rx()._apply_operator(lambda obj, other: obj and other, other)

    def bool(self) -> 'rx':
        """
        Evaluate the truthiness of the current object.

        This method computes the boolean value of the current reactive expression.
        The result is returned as a new reactive expression, allowing the truthiness
        of the object to be used in further reactive operations.

        Returns
        -------
        rx
            A new reactive expression representing the boolean value of the object.

        Examples
        --------
        Create a reactive expression and evaluate its truthiness:

        >>> import param
        >>> rx_value = param.rx(5)
        >>> rx_bool = rx_value.rx.bool()
        >>> rx_bool.rx.value
        True

        Evaluate the truthiness of an empty reactive expression:

        >>> rx_empty = param.rx([])
        >>> rx_bool = rx_empty.rx.bool()
        >>> rx_bool.rx.value
        False
        """
        return self._as_rx()._apply_operator(bool)

    def buffer(self, n) -> 'rx':
        """
        Collect the last ``n`` items emitted by the reactive expression.

        This method creates a new reactive expression that maintains a buffer of the
        most recent ``n`` items emitted by the current reactive expression. As new values
        are emitted, older values are discarded to keep the buffer size constant.

        Parameters
        ----------
        n : int
            The maximum number of items to retain in the buffer.

        Returns
        -------
        rx
            A new reactive expression containing the buffered items as a list.

        Examples
        --------
        Create a reactive expression and buffer the last 3 emitted items:

        >>> import param
        >>> rx_value = param.rx(1)
        >>> rx_buffer = rx_value.rx.buffer(3)
        >>> rx_buffer.rx.value
        [1]

        Emit new values and observe the buffered results:

        >>> rx_value.rx.value = 2
        >>> rx_value.rx.value = 3
        >>> rx_value.rx.value = 4
        >>> rx_buffer.rx.value
        [2, 3, 4]
        """
        items = []
        def collect(new, n):
            items.append(new)
            while len(items) > n:
                items.pop(0)
            return items
        return self._as_rx()._apply_operator(collect, n)

    def in_(self, other) -> 'rx':
        """
        Check if the current object is contained "in" the given operand.

        This method performs a containment check, equivalent to the ``in`` keyword in Python,
        but within a reactive expression. The result is returned as a new reactive expression.

        Parameters
        ----------
        other : any
            The collection or operand to check for containment.

        Returns
        -------
        rx
            A new reactive expression representing the result of the containment check.

        Examples
        --------
        Check if a reactive value is in a list:

        >>> import param
        >>> rx_value = param.rx(2)
        >>> rx_in = rx_value.rx.in_([1, 2, 3])
        >>> rx_in.rx.value
        True

        Check containment with a string:

        >>> rx_char = param.rx('a')
        >>> rx_in = rx_char.rx.in_('alphabet')
        >>> rx_in.rx.value
        True

        Update the reactive value and observe changes:

        >>> rx_char.rx.value = 'c'
        >>> rx_in.rx.value
        False
        """
        return self._as_rx()._apply_operator(operator.contains, other, reverse=True)

    def is_(self, other) -> 'rx':
        """
        Perform a logical "is" comparison with the given operand.

        This method checks if the current object is the same object (i.e., identical in
        memory) as the given operand. The result is returned as a new reactive expression.

        Parameters
        ----------
        other : any
            The operand to compare for object identity.

        Returns
        -------
        rx
            A new reactive expression representing the result of the "is" comparison.

        Examples
        --------
        Check if a reactive value refers to the same object:

        >>> import param
        >>> obj1 = object()
        >>> obj2 = object()
        >>> rx_obj = param.rx(obj1)
        >>> rx_is = rx_obj.rx.is_(obj1)
        >>> rx_is.rx.value
        True

        Compare with a different object:

        >>> rx_is = rx_obj.rx.is_(obj2)
        >>> rx_is.rx.value
        False

        Update the reactive value and re-check identity:

        >>> rx_obj.rx.value = obj2
        >>> rx_is.rx.value
        False
        """
        return self._as_rx()._apply_operator(operator.is_, other)

    def is_not(self, other) -> 'rx':
        """
        Perform a logical "is not" comparison with the given operand.

        This method checks if the current object is not the same object (i.e., not
        identical in memory) as the given operand. The result is returned as a new
        reactive expression.

        Parameters
        ----------
        other : any
            The operand to compare for non-identity.

        Returns
        -------
        rx
            A new reactive expression representing the result of the "is not" comparison.

        Examples
        --------
        Check if a reactive value is not the same object:

        >>> import param
        >>> obj1 = object()
        >>> obj2 = object()
        >>> rx_obj = param.rx(obj1)
        >>> rx_is_not = rx_obj.rx.is_not(obj2)
        >>> rx_is_not.rx.value
        True

        Compare with the same object:

        >>> rx_is_not = rx_obj.rx.is_not(obj1)
        >>> rx_is_not.rx.value
        False

        Update the reactive value and re-check non-identity:

        >>> rx_obj.rx.value = obj2
        >>> rx_is_not.rx.value
        True
        """
        return self._as_rx()._apply_operator(operator.is_not, other)

    def len(self) -> 'rx':
        """
        Return the length of the current object as a reactive expression.

        Since the ``__len__`` method cannot be overloaded reactively, this method
        provides a way to compute the length of the current reactive expression.
        The result is returned as a new reactive expression.

        Returns
        -------
        rx
            A new reactive expression representing the length of the object.

        Examples
        --------
        Compute the length of a reactive list:

        >>> import param
        >>> rx_list = param.rx([1, 2, 3])
        >>> rx_len = rx_list.rx.len()
        >>> rx_len.rx.value
        3

        Update the reactive list and observe the length update:

        >>> rx_list.rx.value = [1, 2, 3, 4, 5]
        >>> rx_len.rx.value
        5

        Compute the length of a reactive string:

        >>> rx_string = param.rx("Hello World")
        >>> rx_len = rx_string.rx.len()
        >>> rx_len.rx.value
        11
        """
        return self._as_rx()._apply_operator(len)

    def map(self, func, /, *args, **kwargs) -> 'rx':
        """
        Apply a function to each item in the reactive collection.

        This method applies a given function to every item in the current reactive
        collection and returns a new reactive expression containing the results.

        Parameters
        ----------
        func : callable
            The function to apply to each item in the collection.
        *args : iterable, optional
            Positional arguments to pass to the function.
        **kwargs : dict, optional
            Keyword arguments to pass to the function.

        Returns
        -------
        rx
            A new reactive expression containing the results of applying ``func``
            to each item.

        Raises
        ------
        TypeError
            If ``func`` is a generator or asynchronous generator function, which
            are not supported.

        Examples
        --------
        Apply a function to double each item in a reactive list:

        >>> import param
        >>> reactive_list = param.rx([1, 2, 3])
        >>> reactive_doubled = reactive_list.rx.map(lambda x: x * 2)
        >>> reactive_doubled.rx.value
        [2, 4, 6]

        Use positional arguments in the function:

        >>> reactive_offset = reactive_list.rx.map(lambda x, offset: x + offset, offset=10)
        >>> reactive_offset.rx.value
        [11, 12, 13]

        Use keyword arguments in the function:

        >>> reactive_power = reactive_list.rx.map(lambda x, power=1: x ** power, power=3)
        >>> reactive_power.rx.value
        [1, 8, 27]
        """
        if inspect.isasyncgenfunction(func) or inspect.isgeneratorfunction(func):
            raise TypeError(
                "Cannot map a generator function. Only regular function "
                "or coroutine functions are permitted."
            )
        if inspect.iscoroutinefunction(func):
            import asyncio
            async def apply(vs, *args, **kwargs):
                return list(await asyncio.gather(*(func(v, *args, **kwargs) for v in vs)))
        else:
            def apply(vs, *args, **kwargs):
                return [func(v, *args, **kwargs) for v in vs]
        return self._as_rx()._apply_operator(apply, *args, **kwargs)

    def not_(self) -> 'rx':
        """
        Perform a logical NOT operation on the current reactive value.

        This method computes the logical negation (``not``) of the current reactive
        expression and returns the result as a new reactive expression.

        Returns
        -------
        rx
            A new reactive expression representing the result of the NOT operation.

        Examples
        --------
        Apply a logical NOT operation to a reactive boolean:

        >>> import param
        >>> rx_bool = param.rx(True)
        >>> rx_not = rx_bool.rx.not_()
        >>> rx_not.rx.value
        False

        Update the reactive value and observe the change:

        >>> rx_bool.rx.value = False
        >>> rx_not.rx.value
        True
        """
        return self._as_rx()._apply_operator(operator.not_)

    def or_(self, other)-> 'rx':
        """
        Perform a logical OR operation with the given operand.

        This method computes a logical OR (``or``) operation between the current
        reactive value and the provided operand. The result is returned as a new
        reactive expression.

        Parameters
        ----------
        other : any
            The operand to combine with the current value using the OR operation.

        Returns
        -------
        rx
            A new reactive expression representing the result of the OR operation.

        Examples
        --------
        Combine two reactive boolean values using OR:

        >>> import param
        >>> rx_bool1 = param.rx(False)
        >>> rx_bool2 = param.rx(True)
        >>> rx_or = rx_bool1.rx.or_(rx_bool2)
        >>> rx_or.rx.value
        True

        Combine a reactive value with a static boolean:

        >>> rx_or_static = rx_bool1.rx.or_(True)
        >>> rx_or_static.rx.value
        True

        Update the reactive value and observe the change:

        >>> rx_bool1.rx.value = True
        >>> rx_or.rx.value
        True
        """
        return self._as_rx()._apply_operator(lambda obj, other: obj or other, other)

    def pipe(self, func, /, *args, **kwargs)-> 'rx':
        """
        Apply a chainable function to the current reactive value.

        This method allows applying a custom function to the current reactive
        expression. The result is returned as a new reactive expression, making
        it possible to create dynamic and chainable pipelines.

        Parameters
        ----------
        func : callable
            The function to apply to the current value.
        *args : iterable, optional
            Positional arguments to pass to the function.
        **kwargs : dict, optional
            Keyword arguments to pass to the function.

        Returns
        -------
        rx
            A new reactive expression representing the result of applying ``func``.

        Examples
        --------
        Apply a custom function to transform a reactive value:

        >>> import param
        >>> rx_value = param.rx(10)
        >>> rx_result = rx_value.rx.pipe(lambda x: x * 2)
        >>> rx_result.rx.value
        20

        Use positional arguments with the function:

        >>> def add(x, y):
        ...     return x + y
        >>> rx_result = rx_value.rx.pipe(add, 5)
        >>> rx_result.rx.value
        15

        Use keyword arguments with the function:

        >>> def multiply(x, factor=1):
        ...     return x * factor
        >>> rx_result = rx_value.rx.pipe(multiply, factor=3)
        >>> rx_result.rx.value
        30
        """
        return self._as_rx()._apply_operator(func, *args, **kwargs)

    def resolve(self, nested=True, recursive=False) -> 'rx':
        """
        Resolve references held by the reactive expression.

        This method resolves references within the reactive expression, replacing
        any references with their actual values. For example, if the expression
        contains a list of reactive parameters, this operation returns a list of
        their resolved values.

        Parameters
        ----------
        nested : bool, optional
            Whether to resolve references within nested objects such as tuples,
            lists, sets, and dictionaries. Default is True.
        recursive : bool, optional
            Whether to recursively resolve references. If a resolved reference
            itself contains further references, this option enables resolving
            them until no references remain. Default is False.

        Returns
        -------
        rx
            A new reactive expression containing the fully resolved values.

        Examples
        --------
        Resolve a simple reactive list of values:

        >>> import param
        >>> rx_list = param.rx([param.rx(1), param.rx(2), param.rx(3)])
        >>> resolved = rx_list.rx.resolve()
        >>> resolved.rx.value
        [1, 2, 3]

        Enable recursive resolution for deeper references:

        >>> rx_nested = param.rx({'key': param.rx([param.rx(10), param.rx(20)])})
        >>> resolved_nested = rx_nested.rx.resolve(recursive=True)
        >>> resolved_nested.rx.value
        {'key': [10, 20]}
        """
        resolver_type = NestedResolver if nested else Resolver
        resolver = resolver_type(object=self._reactive, recursive=recursive)
        return resolver.param.value.rx()

    def updating(self) -> 'rx':
        """
        Return a new expression that indicates whether the current expression is updating.

        This method creates a reactive expression that evaluates to ``True`` while the
        current expression is in the process of updating and ``False`` otherwise. This
        can be useful for tracking or reacting to the update state of an expression,
        such as displaying loading indicators or triggering conditional logic.

        Returns
        -------
        ReactiveExpression
            A reactive expression that is ``True`` while the current expression is updating
            and ``False`` otherwise.

        Examples
        --------
        Create a reactive expression and track its update state:

        >>> import param
        >>> rx_value = param.rx(1)

        Create an updating tracker:

        >>> updating = rx_value.rx.updating()
        >>> updating.rx.value
        False

        Simulate an update and observe the change in the updating tracker:

        >>> rx_value.rx.value = 2
        >>> updating.rx.value  # Becomes True during the update process, then False.
        False
        """
        wrapper = Wrapper(object=False)
        self._watch(lambda e: wrapper.param.update(object=True), precedence=-999)
        self._watch(lambda e: wrapper.param.update(object=False), precedence=999)
        return wrapper.param.object.rx()

    def when(self, *dependencies, initial=Undefined) -> 'rx':
        """
        Create a reactive expression that updates only when specified dependencies change.

        This method creates a new reactive expression that emits the value of the
        current expression only when one of the provided dependencies changes. If
        all dependencies are of type :class:`param.Event` and an initial value is provided,
        the expression will not be evaluated until the first event is triggered.

        Parameters
        ----------
        dependencies : Parameter or reactive expression rx
            Dependencies that trigger an update in the reactive expression.
        initial : object, optional
            A placeholder value that is used until a dependency event is triggered.
            Defaults to ``Undefined``.

        Returns
        -------
        rx
            A reactive expression that updates when the specified dependencies change.

        Examples
        --------
        Use ``.when`` to control when a reactive expression updates:

        >>> import param
        >>> from param import rx
        >>> from time import sleep

        Define an expensive function:

        >>> def expensive_function(a, b):
        ...     print(f'multiplying {a=} and {b=}')
        ...     sleep(1)
        ...     return a * b

        Create reactive values:

        >>> a = rx(1)
        >>> b = rx(2)

        Define a state with an event parameter:

        >>> class State(param.Parameterized):
        ...     submit = param.Event()

        >>> state = State()

        Create a gated reactive expression that only updates when the ``submit`` event is triggered:

        >>> gated_expr = rx(expensive_function)(a, b).rx.when(state.param.submit, initial="Initial Value")
        >>> gated_expr.rx.value
        'Initial Value'

        Trigger the update by setting the `submit` event:

        >>> state.submit = True
        >>> gated_expr.rx.value
        multiplying a=1 and b=2
        2
        """
        deps = [p for d in dependencies for p in resolve_ref(d)]
        is_event = all(isinstance(dep, Event) for dep in deps)
        def eval(*_, evaluated=[]):
            if is_event and initial is not Undefined and not evaluated:
                # Abuse mutable default value to keep track of evaluation state
                evaluated.append(True)
                return initial
            else:
                return self.value
        return bind(eval, *deps).rx()

    def where(self, x, y) -> 'rx':
        """
        Return either ``x`` or ``y`` depending on the current state of the expression.

        This method implements a reactive version of a ternary conditional expression.
        It evaluates the current reactive expression as a condition and returns ``x``
        if the condition is ``True``, or ``y`` if the condition is ``False``.

        Parameters
        ----------
        x : object
            The value to return if the reactive condition evaluates to ``True``.
        y : object
            The value to return if the reactive condition evaluates to ``False``.

        Returns
        -------
        rx
            A reactive expression that evaluates to ``x`` or ``y`` based on the current
            state of the condition.

        Examples
        --------
        Use ``.where`` to implement a reactive conditional:

        >>> import param
        >>> rx_value = param.rx(True)
        >>> rx_result = rx_value.rx.where("Condition is True", "Condition is False")

        Check the result when the condition is ``True``:

        >>> rx_result.rx.value
        'Condition is True'

        Change the reactive condition and observe the updated result:

        >>> rx_value.rx.value = False
        >>> rx_result.rx.value
        'Condition is False'

        Combine ``.where`` with reactive expressions for dynamic updates:

        >>> rx_num = param.rx(10)
        >>> rx_condition = rx_num > 5
        >>> rx_result = rx_condition.rx.where("Above 5", "5 or below")
        >>> rx_result.rx.value
        'Above 5'

        Update the reactive value and see the conditional result change:

        >>> rx_num.rx.value = 3
        >>> rx_result.rx.value
        '5 or below'
        """
        xrefs = resolve_ref(x)
        yrefs = resolve_ref(y)
        if isinstance(self._reactive, rx):
            params = self._reactive._params
        else:
            params = resolve_ref(self._reactive)
        trigger = Trigger(parameters=params)
        if xrefs:
            def trigger_x(*args):
                if self.value:
                    trigger.param.trigger('value')
            bind(trigger_x, *xrefs, watch=True)
        if yrefs:
            def trigger_y(*args):
                if not self.value:
                    trigger.param.trigger('value')
            bind(trigger_y, *yrefs, watch=True)

        def ternary(condition, _):
            return resolve_value(x) if condition else resolve_value(y)
        return bind(ternary, self._reactive, trigger.param.value)

    # Operations to get the output and set the input of an expression

    def set(self, value):
        """
        Set the input of the pipeline to a new value. Equivalent
        to ``.rx.value = value``.

        Parameters
        ----------
        value : object
            The value to set the pipeline input to.
        """
        self.value = value

    @property
    def value(self):
        """
        Get or set the current state of the reactive expression.

        When getting the value it evaluates the reactive expression, resolving all operations
        and dependencies to return the current value. The value reflects the
        latest state of the expression after applying all transformations or updates.

        Returns
        -------
        any
            The current value of the reactive expression, resolved based on the
            operations and dependencies in its pipeline.

        Examples
        --------
        Access the value of a basic reactive expression:

        >>> import param
        >>> rx_value = param.rx(10)
        >>> rx_value.rx.value
        10

        Update the reactive expression and retrieve the updated value:

        >>> rx_value.rx.value = 20
        >>> rx_value.rx.value
        20

        Evaluate a reactive pipeline:

        >>> rx_pipeline = rx_value.rx.pipe(lambda x: x * 2)
        >>> rx_pipeline.rx.value
        40
        """
        if isinstance(self._reactive, rx):
            return self._reactive._resolve()
        elif isinstance(self._reactive, Parameter):
            return getattr(self._reactive.owner, self._reactive.name)
        else:
            return self._reactive()

    @value.setter
    def value(self, new):
        """
        Get or set the current state of the reactive expression.

        When getting the value it evaluates the reactive expression, resolving all operations
        and dependencies to return the current value. The value reflects the
        latest state of the expression after applying all transformations or updates.

        Returns
        -------
        any
            The current value of the reactive expression, resolved based on the
            operations and dependencies in its pipeline.

        Examples
        --------
        Access the value of a basic reactive expression:

        >>> import param
        >>> rx_value = param.rx(10)
        >>> rx_value.rx.value
        10

        Update the reactive expression and retrieve the updated value:

        >>> rx_value.rx.value = 20
        >>> rx_value.rx.value
        20

        Evaluate a reactive pipeline:

        >>> rx_pipeline = rx_value.rx.pipe(lambda x: x * 2)
        >>> rx_pipeline.rx.value
        40
        """
        if isinstance(self._reactive, Parameter):
            raise AttributeError(
                "`Parameter.rx.value = value` is not supported. Cannot override "
                "parameter value."
            )
        elif not isinstance(self._reactive, rx):
            raise AttributeError(
                "`bind(...).rx.value = value` is not supported. Cannot override "
                "the output of a function."
            )
        elif self._reactive._root is not self._reactive:
            raise AttributeError(
                "The value of a derived expression cannot be set. Ensure you "
                "set the value on the root node wrapping a concrete value, e.g.:"
                "\n\n    a = rx(1)\n    b = a + 1\n    a.rx.value = 2\n\n "
                "is valid but you may not set `b.rx.value = 2`."
            )
        if self._reactive._wrapper is None:
            raise AttributeError(
                "Setting the value of a reactive expression is only "
                "supported if it wraps a concrete value. A reactive "
                "expression wrapping a Parameter or another dynamic "
                "reference cannot be updated."
            )
        self._reactive._wrapper.object = resolve_value(new)

    def watch(self, fn=None, onlychanged=True, queued=False, precedence=0):
        """
        Add a callback to observe changes in the reactive expression's output.

        This method allows you to attach a callable function (``fn``) that will be
        invoked whenever the output of the reactive expression changes. The callback
        can be either a regular or asynchronous function. If no callable is provided,
        the expression is eagerly evaluated whenever it updates.

        Parameters
        ----------
        fn : callable or coroutine function, optional
            The function to be called whenever the reactive expression changes.
            For function should accept a single argument, which is the new value
            of the reactive expression. If no function provided, the expression
            is simply evaluated eagerly.

        Raises
        ------
        ValueError
            If ``precedence`` is negative, as negative precedences are reserved
            for internal watchers.

        Examples
        --------
        Attach a regular callback to print the updated value:

        >>> import param
        >>> rx_value = param.rx(10)
        >>> rx_value.rx.watch(lambda v: print(f"Updated value: {v}"))

        Update the reactive value to trigger the callback:

        >>> rx_value.rx.value = 20
        Updated value: 20

        Attach an asynchronous callback:

        >>> import asyncio
        >>> async def async_callback(value):
        ...     await asyncio.sleep(1)
        ...     print(f"Async updated value: {value}")
        >>> rx_value.rx.watch(async_callback)

        Trigger the async callback:

        >>> rx_value.rx.value = 30
        Async updated value: 30  # Printed after a 1-second delay.
        """
        if precedence < 0:
            raise ValueError("User-defined watch callbacks must declare "
                             "a positive precedence. Negative precedences "
                             "are reserved for internal Watchers.")
        self._watch(fn, onlychanged=onlychanged, queued=queued, precedence=precedence)

    def _watch(self, fn=None, onlychanged=True, queued=False, precedence=0):
        def cb(value):
            from .parameterized import async_executor
            if iscoroutinefunction(fn):
                async_executor(partial(fn, value))
            elif fn is not None:
                fn(value)
        bind(cb, self._reactive, watch=True)


def bind(function, *args, watch: bool = False, **kwargs):
    """
    Bind constant values, parameters, bound functions or reactive expressions to a function.

    This function creates a wrapper around the given ``function``, binding some or
    all of its arguments to constant values, :class:`Parameter` objects, or
    reactive expressions. The resulting function automatically reflects updates
    to any bound parameters or reactive expressions, ensuring that its output
    remains up-to-date.

    Similar to :func:`functools.partial`, arguments can also be bound to constants,
    leaving a simple callable object. When ``watch=True``, the function is
    automatically evaluated whenever any bound parameter or reactive expression changes.

    Parameters
    ----------
    function : callable, generator, async generator, or coroutine
        The function or coroutine to bind constant, dynamic, or reactive arguments to.
        It can be:

        - A standard callable (e.g., a regular function).
        - A generator function (producing iterables).
        - An async generator function (producing asynchronous iterables).
        - A coroutine function (producing awaitables).
    *args : object, Parameter, bound function or reactive expression rx
        Positional arguments to bind to the function. These can be constants,
        `param.Parameter` objects, bound functions or reactive expressions.
    watch : bool, optional
        If `True`, the function is automatically evaluated whenever a bound
        parameter or reactive expression changes. Defaults to `False`.
    **kwargs : object, Parameter, bound function or reactive expression rx
        Keyword arguments to bind to the function. These can also be constants,
        `param.Parameter` objects, bound functions or reactive expressions.

    Returns
    -------
    callable, generator, async generator, or coroutine
        A new function with the bound arguments, annotated with all dependencies.
        The function reflects changes to bound parameters or reactive expressions.

    Examples
    --------
    Bind parameters to a function:

    >>> import param
    >>> class Example(param.Parameterized):
    ...     a = param.Number(1)
    ...     b = param.Number(2)
    >>> example = Example()
    >>> def add(a, b):
    ...     return a + b
    >>> bound_add = param.bind(add, example.param.a, example.param.b)
    >>> bound_add()
    3

    Update a parameter and observe the updated result:

    >>> example.a = 5
    >>> bound_add()
    7

    Automatically evaluate the function when bound arguments change:

    >>> bound_watch = param.bind(print, example.param.a, example.param.b, watch=True)
    >>> example.a = 1  # Triggers automatic evaluation
    1 2
    """
    args, kwargs = (
        tuple(transform_reference(arg) for arg in args),
        {key: transform_reference(arg) for key, arg in kwargs.items()}
    )
    dependencies = {}

    # If the wrapped function has a dependency add it
    fn_dep = transform_reference(function)
    if isinstance(fn_dep, Parameter) or hasattr(fn_dep, '_dinfo'):
        dependencies['__fn'] = fn_dep

    # Extract dependencies from args and kwargs
    for i, p in enumerate(args):
        if hasattr(p, '_dinfo'):
            for j, arg in enumerate(p._dinfo['dependencies']):
                dependencies[f'__arg{i}_arg{j}'] = arg
            for kw, kwarg in p._dinfo['kw'].items():
                dependencies[f'__arg{i}_arg_{kw}'] = kwarg
        elif isinstance(p, Parameter):
            dependencies[f'__arg{i}'] = p
    for kw, v in kwargs.items():
        if hasattr(v, '_dinfo'):
            for j, arg in enumerate(v._dinfo['dependencies']):
                dependencies[f'__kwarg_{kw}_arg{j}'] = arg
            for pkw, kwarg in v._dinfo['kw'].items():
                dependencies[f'__kwarg_{kw}_{pkw}'] = kwarg
        elif isinstance(v, Parameter):
            dependencies[kw] = v

    def combine_arguments(wargs, wkwargs, asynchronous=False):
        combined_args = []
        for arg in args:
            if hasattr(arg, '_dinfo'):
                arg = eval_function_with_deps(arg)
            elif isinstance(arg, Parameter):
                arg = getattr(arg.owner, arg.name)
            combined_args.append(arg)
        combined_args += list(wargs)

        combined_kwargs = {}
        for kw, arg in kwargs.items():
            if hasattr(arg, '_dinfo'):
                arg = eval_function_with_deps(arg)
            elif isinstance(arg, Parameter):
                arg = getattr(arg.owner, arg.name)
            combined_kwargs[kw] = arg
        for kw, arg in wkwargs.items():
            if asynchronous:
                if kw.startswith('__arg'):
                    index = kw[5:]
                    if index.isdigit():
                        combined_args[int(index)] = arg
                elif kw.startswith('__kwarg'):
                    substring = kw[8:]
                    if substring in combined_kwargs:
                        combined_kwargs[substring] = arg
                continue
            elif kw.startswith('__arg') or kw.startswith('__kwarg') or kw.startswith('__fn'):
                continue
            combined_kwargs[kw] = arg
        return combined_args, combined_kwargs

    def eval_fn():
        if callable(function):
            fn = function
        else:
            p = transform_reference(function)
            if isinstance(p, Parameter):
                fn = getattr(p.owner, p.name)
            else:
                fn = eval_function_with_deps(p)
        return fn

    if inspect.isgeneratorfunction(function):
        def wrapped(*wargs, **wkwargs):
            combined_args, combined_kwargs = combine_arguments(
                wargs, wkwargs, asynchronous=True
            )
            evaled = eval_fn()(*combined_args, **combined_kwargs)
            for val in evaled:
                yield val
        wrapper_fn = depends(**dependencies, watch=watch)(wrapped)
        wrapped._dinfo = wrapper_fn._dinfo
    elif inspect.isasyncgenfunction(function):
        async def wrapped(*wargs, **wkwargs):
            combined_args, combined_kwargs = combine_arguments(
                wargs, wkwargs, asynchronous=True
            )
            evaled = eval_fn()(*combined_args, **combined_kwargs)
            async for val in evaled:
                yield val
        wrapper_fn = depends(**dependencies, watch=watch)(wrapped)
        wrapped._dinfo = wrapper_fn._dinfo
    elif iscoroutinefunction(function):
        @depends(**dependencies, watch=watch)
        async def wrapped(*wargs, **wkwargs):
            combined_args, combined_kwargs = combine_arguments(
                wargs, wkwargs, asynchronous=True
            )
            evaled = eval_fn()(*combined_args, **combined_kwargs)
            return await evaled
    else:
        @depends(**dependencies, watch=watch)
        def wrapped(*wargs, **wkwargs):
            combined_args, combined_kwargs = combine_arguments(wargs, wkwargs)
            return eval_fn()(*combined_args, **combined_kwargs)
    wrapped.__bound_function__ = function
    wrapped.rx = reactive_ops(wrapped)
    _reactive_display_objs.add(wrapped)
    for name, accessor in _display_accessors.items():
        setattr(wrapped, name, accessor(wrapped))
    return wrapped

# When we only support python >= 3.11 we should exchange 'rx' with Self type annotation below.
# See https://peps.python.org/pep-0673/

class rx:
    """
    A class for creating reactive expressions by wrapping objects.

    The ``rx`` class allows you to wrap objects and operate on them interactively,
    recording any operations applied. These recorded operations form a pipeline
    that can be replayed dynamically when an operand changes. This makes ``rx``
    particularly useful for building reactive workflows, such as real-time data
    processing or dynamic user interfaces.

    Parameters
    ----------
    obj : any
        The object to wrap, such as a number, string, list, or any supported
        data structure.

    References
    ----------
    For more details, see the user guide:
    https://param.holoviz.org/user_guide/Reactive_Expressions.html

    Examples
    --------
    Instantiate :class:`rx` from an object:

    >>> from param import rx
    >>> reactive_float = rx(3.14)

    Perform operations on the reactive object:

    >>> reactive_result = reactive_float * 2
    >>> reactive_result.value
    6.28

    Update the original value and see the updated result:

    >>> reactive_float.value = 1
    >>> reactive_result.rx.value
    2

    Create a reactive list and compute its length reactively:

    >>> reactive_list = rx([1, 2, 3])
    >>> reactive_length = reactive_list.rx.len()
    >>> reactive_length.rx.value
    3
    """

    _accessors: dict[str, Callable[[rx], Any]] = {}

    _display_options: tuple[str] = ()

    _display_handlers: dict[type, tuple[Any, dict[str, Any]]] = {}

    _method_handlers: dict[str, Callable] = {}

    @classmethod
    def register_accessor(
        cls, name: str, accessor: Callable[[rx], Any],
        predicate: Optional[Callable[[Any], bool]] = None
    ):
        """
        Register an accessor that extends ``rx`` with custom behavior.

        Parameters
        ----------
        name: str
          The name of the accessor will be attribute-accessible under.
        accessor: Callable[[rx], any]
          A callable that will return the accessor namespace object
          given the ``rx`` object it is registered on.
        predicate: Callable[[Any], bool] | None

        """
        cls._accessors[name] = (accessor, predicate)

    @classmethod
    def register_display_handler(cls, obj_type, handler, **kwargs):
        """
        Register a display handler for a specific type of object.

        Makes it possible to define custom display options for
        specific objects.

        Parameters
        ----------
        obj_type: type | callable
          The type to register a custom display handler on.
        handler: Viewable | callable
          A Viewable or callable that is given the object to be displayed
          and the custom keyword arguments.
        kwargs: dict[str, Any]
          Additional display options to register for this type.

        """
        cls._display_handlers[obj_type] = (handler, kwargs)

    @classmethod
    def register_method_handler(cls, method, handler):
        """
        Register a handler that is called when a specific method on
        an object is called.
        """
        cls._method_handlers[method] = handler

    def __new__(cls, obj=None, **kwargs):
        wrapper = None
        obj = transform_reference(obj)
        if kwargs.get('fn'):
            # rx._clone codepath
            fn = kwargs.pop('fn')
            wrapper = kwargs.pop('_wrapper', None)
        elif inspect.isgeneratorfunction(obj) or iscoroutinefunction(obj):
            # Resolves generator and coroutine functions lazily
            wrapper = GenWrapper(object=obj)
            fn = bind(lambda obj: obj, wrapper.param.object)
            obj = Undefined
        elif isinstance(obj, (FunctionType, MethodType)) and hasattr(obj, '_dinfo'):
            # Bound functions and methods are resolved on access
            fn = obj
            obj = None
        elif isinstance(obj, Parameter):
            fn = bind(lambda obj: obj, obj)
            obj = getattr(obj.owner, obj.name)
        else:
            # For all other objects wrap them so they can be updated
            # via .rx.value property
            wrapper = Wrapper(object=obj)
            fn = bind(lambda obj: obj, wrapper.param.object)
        inst = super(rx, cls).__new__(cls)
        inst._fn = fn
        inst._shared_obj = kwargs.get('_shared_obj', None if obj is None else [obj])
        inst._wrapper = wrapper
        return inst

    def __init__(
        self, obj=None, operation=None, fn=None, depth=0, method=None, prev=None,
        _shared_obj=None, _current=None, _wrapper=None, _shared=None,  **kwargs
    ):
        # _init is used to prevent to __getattribute__ to execute its
        # specialized code.
        self._init = False
        display_opts = {}
        for _, opts in self._display_handlers.values():
            for k, o in opts.items():
                display_opts[k] = o
        display_opts.update({
            dopt: kwargs.pop(dopt) for dopt in self._display_options + tuple(display_opts)
            if dopt in kwargs
        })
        self._display_opts = display_opts
        self._method = method
        self._operation = operation
        self._depth = depth
        self._dirty = _current is None
        self._dirty_obj = False
        self._current_task = None
        self._error_state = None
        self._current_ = _current
        # _shared is used for branching rx pipelines where we clone the input.
        # Here we store the original shared input, which makes it possible to
        # cache the input value as long as the shared instance does not store
        # a diverging _method accessor.
        self._shared = _shared
        if isinstance(obj, rx) and not prev:
            self._prev = obj
        else:
            self._prev = prev

        # Define special trigger parameter if operation has to be lazily evaluated
        if operation and (iscoroutinefunction(operation['fn']) or inspect.isgeneratorfunction(operation['fn'])):
            self._trigger = Trigger(internal=True)
            self._current_ = Undefined
        else:
            self._trigger = None
        self._root = self._compute_root()
        self._fn_params = self._compute_fn_params()
        self._internal_params = self._compute_params()
        # Filter params that external objects depend on, ensuring
        # that Trigger parameters do not cause double execution
        self._params = [
            p for p in self._internal_params if (not isinstance(p.owner, Trigger) or p.owner.internal)
            or any (p not in self._internal_params for p in p.owner.parameters)
        ]
        self._setup_invalidations(depth)
        self._kwargs = kwargs
        self._rx = reactive_ops(self)
        self._init = True
        for name, accessor in _display_accessors.items():
            setattr(self, name, accessor(self))
        for name, (accessor, predicate) in rx._accessors.items():
            if predicate is None or predicate(self._current):
                setattr(self, name, accessor(self))

    @property
    def rx(self) -> reactive_ops:
        """
        The reactive operations namespace.

        Provides reactive versions of operations that cannot be made reactive through
        operator overloading. This includes operations such as ``.rx.and_`` and ``.rx.bool``.

        References
        ----------
        For more details, see the user guide:
        https://param.holoviz.org/user_guide/Reactive_Expressions.html#special-methods-on-rx

        Examples
        --------
        Create a reactive expression:

        >>> import param
        >>> rx_expression = param.rx(1)

        Retrieve the current value reactively:

        >>> a_value = rx_expression.rx.value

        Use special methods from the reactive ops namespace for reactive operations:

        >>> condition = rx_expression.rx.and_(True)
        >>> piped = rx_expression.rx.pipe(lambda x: x * 2)
        """
        return self._rx

    @property
    def _obj(self):
        if self._shared_obj is None:
            self._obj = eval_function_with_deps(self._fn)
        elif self._root._dirty_obj:
            root = self._root
            root._shared_obj[0] = eval_function_with_deps(root._fn)
            root._dirty_obj = False
        return self._shared_obj[0]

    @property
    def _is_async(self) -> bool:
        if not self._operation:
            return False
        fn = self._operation["fn"]
        return (
            inspect.iscoroutinefunction(fn) or
            inspect.isasyncgenfunction(fn) or
            inspect.isgeneratorfunction(fn)
        )

    @_obj.setter
    def _obj(self, obj):
        if self._shared_obj is None:
            self._shared_obj = [obj]
        else:
            self._shared_obj[0] = obj

    @property
    def _current(self):
        if self._error_state:
            raise self._error_state
        elif self._dirty or self._root._dirty_obj:
            self._resolve()
        return self._current_

    def _compute_root(self):
        if self._prev is None:
            return self
        root = self
        while root._prev is not None:
            root = root._prev
        return root

    def _compute_fn_params(self) -> list[Parameter]:
        if self._fn is None:
            return []

        owner = get_method_owner(self._fn)
        if owner is not None:
            deps = [
                dep.pobj for dep in owner.param.method_dependencies(self._fn.__name__)
            ]
            return deps

        dinfo = getattr(self._fn, '_dinfo', {})
        args = list(dinfo.get('dependencies', []))
        kwargs = list(dinfo.get('kw', {}).values())
        return args + kwargs

    def _compute_params(self) -> list[Parameter]:
        ps = list(self._fn_params)
        if self._trigger:
            ps.append(self._trigger.param.value)

        # Collect parameters on previous objects in chain
        prev = self._prev
        while prev is not None:
            for p in prev._params:
                if p not in ps:
                    ps.append(p)
            prev = prev._prev

        if self._operation is None:
            return ps

        # Accumulate dependencies in args and/or kwargs
        for ref in resolve_ref(self._operation['fn']):
            if ref not in ps:
                ps.append(ref)
        for arg in list(self._operation['args'])+list(self._operation['kwargs'].values()):
            for ref in resolve_ref(arg, recursive=True):
                if ref not in ps:
                    ps.append(ref)

        return ps

    def _setup_invalidations(self, depth: int = 0):
        """
        Since the parameters of the pipeline can change at any time
        we have to invalidate the internal state of the pipeline.
        To handle both invalidations of the inputs of the pipeline
        and the pipeline itself we set up watchers on both.

        1. The first invalidation we have to set up is to re-evaluate
           the function that feeds the pipeline. Only the root node of
           a pipeline has to perform this invalidation because all
           leaf nodes inherit the same shared_obj. This avoids
           evaluating the same function for every branch of the pipeline.
        2. The second invalidation is for the pipeline itself, i.e.
           if any parameter changes we have to notify the pipeline that
           it has to re-evaluate the pipeline. This is done by marking
           the pipeline as `_dirty`. The next time the `_current` value
           is requested the value is resolved by re-executing the
           pipeline.
        """
        if self._fn is not None:
            for _, params in full_groupby(self._fn_params, lambda x: id(x.owner)):
                fps = [p.name for p in params if p in self._root._fn_params]
                if fps:
                    params[0].owner.param._watch(self._invalidate_obj, fps, precedence=-1)
        for _, params in full_groupby(self._internal_params, lambda x: id(x.owner)):
            params[0].owner.param._watch(self._invalidate_current, [p.name for p in params], precedence=-1)

    def _invalidate_current(self, *events):
        if all(event.obj is self._trigger for event in events):
            return
        self._dirty = True
        self._error_state = None

    def _invalidate_obj(self, *events):
        self._root._dirty_obj = True
        self._error_state = None

    async def _resolve_async(self, obj = None):
        import asyncio
        self._current_task = task = asyncio.current_task()
        if obj is None:
            if self._shared._current_task:
                await self._shared._current_task
            self._current_ = self._shared.rx.value
            self._trigger.param.trigger('value')
        elif inspect.isasyncgen(obj):
            async for val in obj:
                if self._current_task is not task:
                    break
                self._current_ = val
                self._trigger.param.trigger('value')
        else:
            value = await obj
            if self._current_task is task:
                self._current_ = value
                self._trigger.param.trigger('value')

    def _lazy_resolve(self, obj = None):
        from .parameterized import async_executor
        if inspect.isgenerator(obj):
            obj = _to_async_gen(obj)
        async_executor(partial(self._resolve_async, obj))

    def _resolve(self):
        if self._error_state:
            raise self._error_state
        elif self._dirty or self._root._dirty_obj:
            try:
                obj = self._obj if self._prev is None else self._prev._resolve()
                if obj is Skip or obj is Undefined:
                    self._current_ = Undefined
                    raise Skip
                elif (
                    self._shared is not None and
                    self._method is None and
                    self._shared._method is None
                ):
                    # If this rx is cloned from an shared input then we make use
                    # of the shared.rx.value to ensure branching pipelines do
                    # not have to recompute the inputs multiple times.
                    if self._is_async:
                        self._shared.rx.value # trigger async resolve
                        self._lazy_resolve()
                    else:
                        self._current_ = self._shared.rx.value
                    raise Skip
                operation = self._operation
                if operation:
                    obj = self._eval_operation(obj, operation)
                    if self._is_async:
                        self._lazy_resolve(obj)
                        obj = Skip
                    if obj is Skip:
                        raise Skip
            except Skip:
                self._dirty = False
                return self._current_
            except Exception as e:
                self._error_state = e
                raise e
            self._current_ = current = obj
        else:
            current = self._current_
        self._dirty = False
        if self._method:
            # E.g. `pi = dfi.A` leads to `pi._method` equal to `'A'`.
            current = getattr(current, self._method, current)
        if hasattr(current, '__call__'):
            self.__call__.__func__.__doc__ = self.__call__.__doc__
        return current

    def _transform_output(self, obj):
        """Apply custom display handlers before their output."""
        applies = False
        for predicate, (handler, opts) in self._display_handlers.items():
            display_opts = {
                k: v for k, v in self._display_opts.items() if k in opts
            }
            display_opts.update(self._kwargs)
            try:
                applies = predicate(obj, **display_opts)
            except TypeError:
                applies = predicate(obj)
            if applies:
                new = handler(obj, **display_opts)
                if new is not obj:
                    return new
        return obj

    @property
    def _callback(self):
        params = self._params
        def evaluate(*args, **kwargs):
            out = self._current
            if self._method:
                out = getattr(out, self._method)
            return self._transform_output(out)
        if params:
            return bind(evaluate, *params)
        return evaluate

    def _clone(self, operation=None, copy=False, **kwargs) -> 'rx':
        operation = operation or self._operation
        depth = self._depth + 1
        if copy:
            kwargs = dict(
                self._kwargs, _current=self._current, method=self._method,
                prev=self._prev, _shared=self, **kwargs
            )
        else:
            kwargs = dict(prev=self, **dict(self._kwargs, **kwargs))
        kwargs = dict(self._display_opts, **kwargs)
        return type(self)(
            self._obj, operation=operation, depth=depth, fn=self._fn,
            _shared_obj=self._shared_obj, _wrapper=self._wrapper,
            **kwargs
        )

    def __dir__(self):
        current = self._current
        if self._method:
            current = getattr(current, self._method)
        extras = {attr for attr in dir(current) if not attr.startswith('_')}
        try:
            return sorted(set(super().__dir__()) | extras)
        except Exception:
            return sorted(set(dir(type(self))) | set(self.__dict__) | extras)

    def _resolve_accessor(self):
        if not self._method:
            # No method is yet set, as in `dfi.A`, so return a copied clone.
            return self._clone(copy=True)
        # This is executed when one runs e.g. `dfi.A > 1`, in which case after
        # dfi.A the _method 'A' is set (in __getattribute__) which allows
        # _resolve_accessor to record the attribute access as an operation.
        operation = {
            'fn': getattr,
            'args': (self._method,),
            'kwargs': {},
            'reverse': False
        }
        self._method = None
        return self._clone(operation)

    def __getattribute__(self, name):
        self_dict = super().__getattribute__('__dict__')
        if not self_dict.get('_init') or name == 'rx' or name.startswith('_'):
            return super().__getattribute__(name)

        current = self_dict['_current_']
        dirty = self_dict['_dirty']
        if dirty:
            self._resolve()
            current = self_dict['_current_']

        method = self_dict['_method']
        if method:
            current = getattr(current, method)
        # Getting all the public attributes available on the current object,
        # e.g. `sum`, `head`, etc.
        extras = [d for d in dir(current) if not d.startswith('_')]
        if (name in extras or current is Undefined) and name not in super().__dir__():
            new = self._resolve_accessor()
            # Setting the method name for a potential use later by e.g. an
            # operator or method, as in `dfi.A > 2`. or `dfi.A.max()`
            new._method = name
            try:
                new.__doc__ = getattr(current, name).__doc__
            except Exception:
                pass
            return new
        return super().__getattribute__(name)

    def __call__(self, *args, **kwargs):
        new = self._clone(copy=True)
        method = new._method or '__call__'
        if method == '__call__' and self._depth == 0 and not hasattr(self._current, '__call__'):
            return self.set_display(*args, **kwargs)

        if method in rx._method_handlers:
            handler = rx._method_handlers[method]
            method = handler(self)
        new._method = None
        kwargs = dict(kwargs)
        operation = {
            'fn': method,
            'args': args,
            'kwargs': kwargs,
            'reverse': False
        }
        return new._clone(operation)

    #----------------------------------------------------------------
    # rx pipeline APIs
    #----------------------------------------------------------------

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        new = self._resolve_accessor()
        operation = {
            'fn': getattr(ufunc, method),
            'args': args[1:],
            'kwargs': kwargs,
            'reverse': False
        }
        return new._clone(operation)

    def _apply_operator(self, operator, *args, reverse=False, **kwargs) -> 'rx':
        new = self._resolve_accessor()
        operation = {
            'fn': operator,
            'args': args,
            'kwargs': kwargs,
            'reverse': reverse
        }
        return new._clone(operation)

    # Builtin functions

    def __abs__(self):
        return self._apply_operator(abs)

    def __str__(self):
        return self._apply_operator(str)

    def __round__(self, ndigits=None):
        args = () if ndigits is None else (ndigits,)
        return self._apply_operator(round, *args)

    # Unary operators
    def __ceil__(self):
        return self._apply_operator(math.ceil)
    def __floor__(self):
        return self._apply_operator(math.floor)
    def __invert__(self):
        return self._apply_operator(operator.inv)
    def __neg__(self):
        return self._apply_operator(operator.neg)
    def __pos__(self):
        return self._apply_operator(operator.pos)
    def __trunc__(self):
        return self._apply_operator(math.trunc)

    # Binary operators
    def __add__(self, other):
        return self._apply_operator(operator.add, other)
    def __and__(self, other):
        return self._apply_operator(operator.and_, other)
    def __contains_(self, other):
        return self._apply_operator(operator.contains, other)
    def __divmod__(self, other):
        return self._apply_operator(divmod, other)
    def __eq__(self, other):
        return self._apply_operator(operator.eq, other)
    def __floordiv__(self, other):
        return self._apply_operator(operator.floordiv, other)
    def __ge__(self, other):
        return self._apply_operator(operator.ge, other)
    def __gt__(self, other):
        return self._apply_operator(operator.gt, other)
    def __le__(self, other):
        return self._apply_operator(operator.le, other)
    def __lt__(self, other):
        return self._apply_operator(operator.lt, other)
    def __lshift__(self, other):
        return self._apply_operator(operator.lshift, other)
    def __matmul__(self, other):
        return self._apply_operator(operator.matmul, other)
    def __mod__(self, other):
        return self._apply_operator(operator.mod, other)
    def __mul__(self, other):
        return self._apply_operator(operator.mul, other)
    def __ne__(self, other):
        return self._apply_operator(operator.ne, other)
    def __or__(self, other):
        return self._apply_operator(operator.or_, other)
    def __rshift__(self, other):
        return self._apply_operator(operator.rshift, other)
    def __pow__(self, other):
        return self._apply_operator(operator.pow, other)
    def __sub__(self, other):
        return self._apply_operator(operator.sub, other)
    def __truediv__(self, other):
        return self._apply_operator(operator.truediv, other)
    def __xor__(self, other):
        return self._apply_operator(operator.xor, other)

    # Reverse binary operators
    def __radd__(self, other):
        return self._apply_operator(operator.add, other, reverse=True)
    def __rand__(self, other):
        return self._apply_operator(operator.and_, other, reverse=True)
    def __rdiv__(self, other):
        return self._apply_operator(operator.div, other, reverse=True)
    def __rdivmod__(self, other):
        return self._apply_operator(divmod, other, reverse=True)
    def __rfloordiv__(self, other):
        return self._apply_operator(operator.floordiv, other, reverse=True)
    def __rlshift__(self, other):
        return self._apply_operator(operator.rlshift, other)
    def __rmod__(self, other):
        return self._apply_operator(operator.mod, other, reverse=True)
    def __rmul__(self, other):
        return self._apply_operator(operator.mul, other, reverse=True)
    def __ror__(self, other):
        return self._apply_operator(operator.or_, other, reverse=True)
    def __rpow__(self, other):
        return self._apply_operator(operator.pow, other, reverse=True)
    def __rrshift__(self, other):
        return self._apply_operator(operator.rrshift, other)
    def __rsub__(self, other):
        return self._apply_operator(operator.sub, other, reverse=True)
    def __rtruediv__(self, other):
        return self._apply_operator(operator.truediv, other, reverse=True)
    def __rxor__(self, other):
        return self._apply_operator(operator.xor, other, reverse=True)

    def __getitem__(self, other):
        return self._apply_operator(operator.getitem, other)

    def __iter__(self):
        if isinstance(self._current, Iterator):
            while True:
                try:
                    new = self._apply_operator(next)
                    new.rx.value
                except RuntimeError:
                    break
                yield new
            return
        elif not isinstance(self._current, Iterable):
            raise TypeError(f'cannot unpack non-iterable {type(self._current).__name__} object.')
        items = self._apply_operator(list)
        for i in range(len(self._current)):
            yield items[i]

    def __bool__(self):
        # Implemented otherwise truth value testing (e.g. if rx: ...)
        # defers to __len__ which raises an error.
        return True

    def __len__(self):
        raise TypeError(
            'len(<rx_obj>) is not supported. Use `<rx_obj>.rx.len()` to '
            'obtain the length as a reactive expression, or '
            '`len(<rx_obj>.rx.value)` to obtain the length of the underlying '
            'expression value.'
        )

    def _eval_operation(self, obj, operation):
        fn, args, kwargs = operation['fn'], operation['args'], operation['kwargs']
        resolved_args = []
        for arg in args:
            val = resolve_value(arg)
            if val is Skip or val is Undefined:
                raise Skip
            resolved_args.append(val)
        resolved_kwargs = {}
        for k, arg in kwargs.items():
            val = resolve_value(arg)
            if val is Skip or val is Undefined:
                raise Skip
            resolved_kwargs[k] = val
        if isinstance(fn, str):
            obj = getattr(obj, fn)(*resolved_args, **resolved_kwargs)
        elif operation.get('reverse'):
            obj = fn(resolved_args[0], obj, *resolved_args[1:], **resolved_kwargs)
        else:
            obj = fn(obj, *resolved_args, **resolved_kwargs)
        return obj

    def __setattr__(self, name, value):
        # Setting value instead of rx.value is a common user mistake.
        # They are more but we don't want to restrict __setattr__ too much
        # so only catch value, for now.
        if name == "value":
            raise AttributeError(
                "'rx' has no attribute 'value', try "
                "'<reactive_expr>.rx.value = <val>'."
            )
        super().__setattr__(name, value)


def _rx_transform(obj):
    if not isinstance(obj, rx):
        return obj
    return bind(lambda *_: obj.rx.value, *obj._params)

register_reference_transform(_rx_transform)
